#include "dlist.h"
#include "exc.h"
#include "extern-c++.h"

extern "C" {

void check_bytedata_ptr(GENERIC*);
void check_lonely_ptr(GENERIC*);
void update_ptr(GENERIC*);

}

//////////////////////
// CLASS B_DL_Node //
////////////////////


B_DL_Node:: ~B_DL_Node() {
  //  cerr << "B_DL_Node:: ~B_DL_Node()\n";
  B_DL_Node *p, *n;
  B_DL_Node *node=this;

  p = node->prev1_d;
  n = node->next1_d;
  if ( p && n ) {
    p->next1_d=n;
    n->prev1_d=p;
  }
  else {
    if (p) p->next1_d=NULL;
    if (n) n->prev1_d=NULL;
  }

}

void B_DL_Node:: insert_node(B_DL_Node* p, B_DL_Node* n){ 
  if (p) { 
    p->next1_d=this; 
    prev1_d=p;
  }
  else {
    prev1_d=NULL;
  };
  if (n) {
    n->prev1_d=this; 
    next1_d=n;
  }
  else {
    next1_d=NULL;
  }
}

B_DL_Node* B_DL_Node:: remove_node1(void) {
  B_DL_Node *p, *n;

  B_DL_Node *node=this;
  UNEXPECTED( (this->header_node()),
	     "cannot remove header_node in B_DL_Node::remove_node1" )
  p = node->prev1_d;
  n = node->next1_d;
  if ( p && n ) {
    p->next1_d=n;
    n->prev1_d=p;
  }
  else {
    if (p) p->next1_d=NULL;
    if (n) n->prev1_d=NULL;
  }
  node->next1_d=node->prev1_d=NULL;
  return node;

}

void B_DL_Node:: check_data() {
  //  cerr << "DL_Node ";

  if (pass==1) {
    if (VALID_LIFE_ADDRESS(next1_d))
      if (next1_d->header_node()) { 
	check_lonely_ptr((GENERIC*)(&next1_d));
      }      
      else {
	check_bytedata_ptr((GENERIC*)(&next1_d));
      }
    
    if (VALID_LIFE_ADDRESS(prev1_d))
      if (prev1_d->header_node()) { 
	check_lonely_ptr((GENERIC*)(&prev1_d));
      }      
      else {
	check_bytedata_ptr((GENERIC*)(&prev1_d));
      }
  }


  if (pass==2) {
    if (VALID_LIFE_ADDRESS(next1_d))
      update_ptr((GENERIC*)(&next1_d));

    if (VALID_LIFE_ADDRESS(prev1_d))
      update_ptr((GENERIC*)(&prev1_d));
  }

  //  cerr << "DL_Node over\n";
}

ostream& operator<<( ostream & os, const B_DL_Node& n ) {
  n.output_data(os);
  return os;
}

////////////
// DList //
//////////
DList:: DList() {first_d.next1_d=&last_d; last_d.prev1_d=&first_d;};

DList:: ~DList(){
}

void DList:: append(B_DL_Node* node) {

  last_d.prepend1(node);

}

void DList:: prepend(B_DL_Node* node) {
  first_d.append1(node);
}

B_DL_Node* DList:: remove_a(void) {
  B_DL_Node* first = first_d.next1_d;

  return first->header_node() ? (B_DL_Node*)NULL : first->remove1();

}

B_DL_Node* DList:: remove_e(void) {
  B_DL_Node* last = last_d.prev1_d;

  return last->header_node() ? (B_DL_Node*)NULL : last->remove1();

}


void DList:: delete_a(void) {
  delete remove_a();
}

void DList:: delete_e(void) {
  delete remove_e();
}

void DList:: output_data(ostream& os)const{

  os << "dlist header: "; 
  os << "[";
  OUTPUT(&first_d, first_d.next1_d, first_d.prev1_d);
  B_DL_Node* np = first();
  if (np) {
    np->output_data(os); 
    np = np->next1();
  }
  for ( ; np; np=np->next1() ) {
    os << ", ";
    np->output_data(os);
  };
  OUTPUT(&last_d, last_d.next1_d, last_d.prev1_d);
  os << "]";

}

int DList:: length() {
  int s=0;
  B_DL_Node* np = first();
  while(np) {
    s++;
    np=np->next1();
  }
  return s;
}

void DList:: check_data() {
  //  cerr << "DList ";

  if (pass==1) {
    if (VALID_LIFE_ADDRESS(first_d.next1_d))
      
      if (first_d.next1_d->header_node()) {
	check_lonely_ptr((GENERIC*)(&(first_d.next1_d)));
      }
      else {
	check_bytedata_ptr((GENERIC*)(&(first_d.next1_d)));
      }
    
    if (VALID_LIFE_ADDRESS(last_d.prev1_d))
      if (last_d.prev1_d->header_node()) {
	check_lonely_ptr((GENERIC*)(&(last_d.prev1_d)));
      }
      else {
	check_bytedata_ptr((GENERIC*)(&(last_d.prev1_d)));
      }
  }

  if (pass==2) {
    if (VALID_LIFE_ADDRESS(first_d.next1_d))
      update_ptr((GENERIC*)(&(first_d.next1_d)));

    if (VALID_LIFE_ADDRESS(last_d.prev1_d))
      update_ptr((GENERIC*)(&(last_d.prev1_d)));
  }

  //  cerr << "DList over ";
}


ostream& operator<<( ostream & os, const DList& list ) {
  list.output_data(os);
  return os;
}


