#include <iostream.h>
#include <unistd.h>
#include <errno.h>
#include <sys/times.h>

#include "file.h"
#include "exc.h"

extern "C" {

}

////////////////
// CLASS File //
///////////////
File:: File(const char* name, int flags, int mode): 
                   file_name(name?new char[strlen(name)+1]:(char*)NULL),
		   state_d(0)
{
  UNEXPECTED( (!file_name),
	     "unexpected null string for file_name in File::File()" );
  strcpy(file_name,name);
  handle_d = ::open(name, flags, mode);
  if ( handle_d == -1 ) {
    dberror << "could not open file " << file_name;
    throw_error();
  }
}

File:: ~File() {
  if (is_open()) ::close(handle_d);
  delete[] file_name;
}


void File:: open(int flags, int mode) {
  if (!is_open()) {
    handle_d = ::open(file_name, flags, mode);
    if ( handle_d == -1 ) {
      dberror << "could not open file " << file_name;
      throw_error(); 
    }
  }
}

void File:: close() {
  if (is_open()) {
    ::close(handle_d);
    handle_d=-1;
  }
}

float File:: read_time=0;
long File:: times_read_called=0;

void File:: read_file(char* buf, int size) const {
  if (is_open()) {
    struct tms read_start_time, read_end_time;
    times(&read_start_time);
    if ( ::read(handle_d, buf, size) == -1 ) {
      dberror << "could not read file -- " << file_name;
      throw_error();
    }
    times(&read_end_time);
    float time_taken = (read_end_time.tms_stime - 
			read_start_time.tms_stime);
    read_time+=time_taken;
    times_read_called++;
  }
  else {
    dberror << "could not read file, invalid file handle for file -- " 
	    << file_name;
    throw_error();
  }
}

float File:: write_time=0;
long File:: times_write_called=0;

void File:: write_file(char* buf, int size) const {
  if (is_open()) {
    struct tms write_start_time, write_end_time;
    times(&write_start_time);
    if ( ::write(handle_d, buf, size) == -1 ) {
      dberror << "could not write file " << file_name;
      throw_error();
    }
    times(&write_end_time);
    float time_taken = (write_end_time.tms_stime - 
			write_start_time.tms_stime);
    write_time+=time_taken;
    times_write_called++;
  }
  else {
    dberror << "could not write file, invalid file handle for file " 
	    << file_name;
    throw_error(); 
  }
}

off_t File:: seek_loc(off_t offset, int whence) const {
  if (is_open()) {
    off_t loc = ::lseek( handle_d, offset, whence);
    if ( loc == -1 ) {
      dberror << "seek error in file " << file_name;
      throw_error(); 
    }
    return loc;
  }
  else {
    dberror << "seek error in file, invalid file handle for file " 
	    << file_name;
    throw_error();
  }
}

off_t File:: seek_end() const {
  return seek_loc(0, SEEK_END);
}

void File:: output_data(ostream& os)const{
  os << "*** File: " << "\"" << file_name << "\"";
}

ostream& operator<<( ostream & os, const File& file){
  file.output_data(os);
  return os;
}


