#include "flist.h"
#include "exc.h"
#include "extern-c++.h"


extern "C" {

void check_bytedata_ptr(GENERIC*);
void check_lonely_ptr(GENERIC*);
void update_ptr(GENERIC*);

}

B_FL_Node:: ~B_FL_Node() {
  //  cerr << "B_FL_Node:: ~B_FL_Node()\n";
  B_FL_Node *p, *n;
  B_FL_Node *node=this;

  p = node->prev2_d;
  n = node->next2_d;
  if ( p && n ) {
    p->next2_d=n;
    n->prev2_d=p;
  }
  else {
    if (p) p->next2_d=NULL;
    if (n) n->prev2_d=NULL;
  }

}

void B_FL_Node:: insert_node(B_FL_Node* p, B_FL_Node* n){ 
  if (p) { 
    p->next2_d=this; 
    prev2_d=p;
  }
  else {
    prev2_d=NULL;
  };
  if (n) {
    n->prev2_d=this; 
    next2_d=n;
  }
  else {
    next2_d=NULL;
  }
}

B_FL_Node* B_FL_Node:: remove_node2(void) {
  B_FL_Node *p, *n;

  B_FL_Node *node=this;
  UNEXPECTED( (this->header_node()),
	     "cannot remove header_node in B_FL_Node::remove_node" )
  p = node->prev2_d;
  n = node->next2_d;
  if ( p && n ) {
    p->next2_d=n;
    n->prev2_d=p;
  }
  else {
    if (p) p->next2_d=NULL;
    if (n) n->prev2_d=NULL;
  }
  node->next2_d=node->prev2_d=NULL;

  return node;

}

void B_FL_Node:: check_data() {
  B_DL_Node:: check_data();
  //  cerr << "FL_Node ";

  if (pass==1) {
    if (VALID_LIFE_ADDRESS(next2_d))
      if (next2_d->header_node()) { 
	check_lonely_ptr((GENERIC*)(&next2_d));
      }      
      else {
	check_bytedata_ptr((GENERIC*)(&next2_d));
      }
    
    if (VALID_LIFE_ADDRESS(prev2_d))
      if (prev2_d->header_node()) { 
	check_lonely_ptr((GENERIC*)(&prev2_d));
      }      
      else {
	check_bytedata_ptr((GENERIC*)(&prev2_d));
      }
  }

  if (pass==2) {
    if (VALID_LIFE_ADDRESS(next2_d))
      update_ptr((GENERIC*)(&next2_d));

    if (VALID_LIFE_ADDRESS(prev2_d))
      update_ptr((GENERIC*)(&prev2_d));
  }

  //  cerr << "FL_Node over\n";
}

////////////
// FList //
//////////
FList:: ~FList(){}

void FList:: append(B_FL_Node* node) {

  last_d.prepend2(node);

}

void FList:: prepend(B_FL_Node* node) {
  first_d.append2(node);
}

B_FL_Node* FList:: remove_a(void) {
  B_FL_Node* f = first();

  return f->header_node() ? (B_FL_Node*)NULL : 
                                (B_FL_Node*)f->remove2();

}

B_FL_Node* FList:: remove_e(void) {
  B_FL_Node* l = last();

  return l->header_node() ? (B_FL_Node*)NULL : 
                               (B_FL_Node*)l->remove2();

}


void FList:: delete_a(void) {
  delete remove_a();
}

void FList:: delete_e(void) {
  delete remove_e();
}

void FList:: check_data() {
  //  cerr << "FList ";

  if (pass==1) {
    if (VALID_LIFE_ADDRESS(first_d.next2_d))
      if (first_d.next2_d->header_node()) {
	check_lonely_ptr((GENERIC*)(&(first_d.next2_d)));
      }
      else {
	check_bytedata_ptr((GENERIC*)(&(first_d.next2_d)));
      }
    
    if (VALID_LIFE_ADDRESS(last_d.prev2_d))
      if (last_d.prev2_d->header_node()) {
	check_lonely_ptr((GENERIC*)(&(last_d.prev2_d)));
      }
      else {
	check_bytedata_ptr((GENERIC*)(&(last_d.prev2_d)));
      }
  }

  if (pass==2) {
    if (VALID_LIFE_ADDRESS(first_d.next2_d))
      update_ptr((GENERIC*)(&(first_d.next2_d)));

    if (VALID_LIFE_ADDRESS(last_d.prev2_d))
      update_ptr((GENERIC*)(&(last_d.prev2_d)));
  }

  //  cerr << "FList over ";
}

void FList:: output_data(ostream& os)const{
  os << "flist header: "; 
  os << "[";
  OUTPUT(&first_d, first_d.next2_d, first_d.prev2_d);
  B_FL_Node* np = first();
  if (np) {
    np->output_data(os); 
    np = np->next2();
  }
  for ( ; np; np=np->next2() ) {
    os << ", ";
    np->output_data(os);
  };
  OUTPUT(&last_d, last_d.next2_d, last_d.prev2_d);
  os << "]";

}

int FList:: length() {
  int s=0;
  B_FL_Node* np = first();
  while(np) {
    s++;
    np=np->next2();
  }
  return s;
}

ostream& operator<<( ostream & os, const FList& list ) {
  list.output_data(os);
  return os;
}


