#ifndef hash_hdr
#define hash_hdr

#include "dlist.h"
#include "exc.h"

////////////////////////
// CLASS B_HashTable //
//////////////////////
class B_HashTable {

protected:
  int size_d;			// array size of hash table
  DList* table_d;
  

public:
  B_HashTable(int);
  ~B_HashTable();

  friend ostream& operator<<(ostream&, const B_HashTable& ht);

protected:
  int capacity(int);
  void output_data(ostream& os)const;

protected:
  static const int num_of_primes;
  static const int hash_primes[];

private:
  B_HashTable(const B_HashTable&);
  void operator=(const B_HashTable&);
};

//////////////////////
// CLASS HashTable //
////////////////////
template<class HashKey, class HashNode>
class HashTable:public B_HashTable {
 
private:
  int i_pos;
  HashNode* i_node;

public:
  HashTable(int  i = 5):B_HashTable(i),i_pos(-1),i_node(NULL) {}

  HashNode* fetch(const HashKey& key) {
    int index=key.xkey() % size_d;

    HashNode* item= (HashNode*) table_d[index].first();
    while (item) {
      if (item->key() == key) break;
      item = (HashNode*) item->next1();
    }
    return item;
  }

  void insert(HashNode* n) {
    int index=n->key().xkey() % size_d;

    UNEXPECTED( ( fetch(n->key()) ),
	       "rehashing node in HashTable:: insert()" );

    table_d[index].prepend(n);
  }

  HashNode* next() {
    if (i_node) i_node = (HashNode*)i_node->next1();
    while (!i_node) {
      i_pos++;
      if (i_pos<size_d) {
	i_node = (HashNode*)table_d[i_pos].first();
      }
      else {
	i_pos = -1;
	i_node = NULL;
	break;
      }
    }
    return i_node;
  }

  void init_iterator() {
    i_pos=-1;
    i_node=NULL;
  }

};


#endif
