#ifndef hash_mat2_hdr
#define hash_mat2_hdr


#include "dlist.h"
#include "exc.h"

//////////////////////////
// CLASS B_HashMatrix2 //
////////////////////////
class B_HashMatrix2 {

protected:
  static const int num_of_primes;
  static const int hash_primes[];
  int x_size_d;			
  int y_size_d;
  DList** table_d;
  
public:
  B_HashMatrix2(int, int);
  ~B_HashMatrix2();

  void usage(ostream& os)const;
  friend ostream& operator<<(ostream& os, const B_HashMatrix2& hm);
protected:
  int capacity(int);
  void output_data(ostream & os)const;

private:
  B_HashMatrix2(const B_HashMatrix2&);
  void operator=(const B_HashMatrix2&);
};

////////////////////////
// CLASS HashMatrix2 //
//////////////////////
template<class HashKey, class HashNode>
class HashMatrix2:public B_HashMatrix2 {
 
private:
  int ix_pos, iy_pos;
  HashNode* i_node;

public:
  HashMatrix2(int  i = 5, int j = 13):B_HashMatrix2(i,j) {}

  HashNode* fetch(const HashKey& key) {
    int xindex=key.xkey() % x_size_d;
    int yindex=key.ykey() % y_size_d;

    HashNode* item= (HashNode*) table_d[xindex][yindex].first();
    while (item) {
      if (item->key() == key) break;
      item = (HashNode*) item->next1();
    }
    return item;
  }

  void insert(HashNode* n) {
    int xindex=n->key().xkey() % x_size_d;
    int yindex=n->key().ykey() % y_size_d;

    UNEXPECTED( ( fetch(n->key()) ),
	       "rehashing node in HashMatrix2:: insert()" );
    table_d[xindex][yindex].prepend(n);
  }

  HashNode* next() {
    if (i_node) i_node = (HashNode*)i_node->next1();
    while (!i_node) {
      iy_pos++;
      if (iy_pos >= y_size_d) {
	iy_pos=0;
	ix_pos++;
	if (ix_pos >= x_size_d) {
	  ix_pos=0;
	  iy_pos=-1;
	  i_node=NULL;
	  break;
	}
      }
      i_node=(HashNode*) table_d[ix_pos][iy_pos].first();
    }
    return i_node;
  }

  void init_iterator() {
    ix_pos=0;
    iy_pos=-1;
    i_node=NULL;
  }

  void check_data() {
    //    cerr << "hash2 ";
    for(int i=0; i < x_size_d; i++) {
      for(int j=0; j < y_size_d; j++){
        table_d[i][j].check_data();
      }
    }
    //    cerr << "hash2 over ";
  }

};

#endif
