#ifndef hash_mat3_hdr
#define hash_mat3_hdr

#include <iostream.h>
#include "dlist.h"

//////////////////////////
// CLASS B_HashMatrix3 //
////////////////////////
class B_HashMatrix3 {

protected:
  static const int num_of_primes;
  static const int hash_primes[];
  int x_size_d;			
  int y_size_d;
  int z_size_d;
  DList*** table_d;
  
public:
  B_HashMatrix3(int, int, int);
  ~B_HashMatrix3();

  void usage(ostream & os)const;
  friend ostream& operator<<(ostream& os, const B_HashMatrix3& hm);

protected:
  int capacity(int);
  void output_data(ostream & os)const;

private:
  B_HashMatrix3(const B_HashMatrix3&);
  void operator=(const B_HashMatrix3&);

};

///////////////////////
// CLASS HashMatrix3 //
/////////////////////
template<class HashKey, class HashNode>
class HashMatrix3:public B_HashMatrix3 {
 
private:
  int ix_pos, iy_pos, iz_pos;
  HashNode* i_node;

public:
  HashMatrix3(int  i = 5, int j = 13, int k = 7):B_HashMatrix3(i,j,k),
          ix_pos(0),iy_pos(0),iz_pos(-1),i_node(NULL) {}

  HashNode* fetch(const HashKey& key) {
    int xindex=key.xkey() % x_size_d;
    int yindex=key.ykey() % y_size_d;
    int zindex=key.zkey() % z_size_d;

    HashNode* item= 
      (HashNode*) table_d[xindex][yindex][zindex].first();
    while (item) {
      if (item->key() == key) break;
      item = (HashNode*) item->next1();
    }
    return item;
  }

  void insert(HashNode* n) {
    int xindex=n->key().xkey() % x_size_d;
    int yindex=n->key().ykey() % y_size_d;
    int zindex=n->key().zkey() % z_size_d;
    

//    UNEXPECTED( ( fetch(n->key()) ),
//	       "rehashing node in HashMatrix3:: insert()" );

    table_d[xindex][yindex][zindex].prepend(n);
  }

  HashNode* next() {
    if (i_node) i_node = (HashNode*)i_node->next1();
    while (!i_node) {
      iz_pos++;
      if (iz_pos >= z_size_d) {
	iz_pos=0;
	iy_pos++;
	if (iy_pos >= y_size_d) {
	  iy_pos=0;
	  ix_pos++;
	  if (ix_pos >= x_size_d) {
	    ix_pos=0;
	    iy_pos=0;
	    iz_pos=-1;
	    i_node=NULL;
	    break;
	  }
	}
      }				// if (iz_pos >= z_size_d)   
      i_node=(HashNode*) table_d[ix_pos][iy_pos][iz_pos].first();
    }
    return i_node;
  }

  void init_iterator() {
    ix_pos=0;
    iy_pos=0;
    iz_pos=-1;
    i_node=NULL;
  }

  void check_data() {
    //    cerr << "hash3 ";
    for(int i=0; i < x_size_d; i++) {
      for(int j=0; j < y_size_d; j++){
        for(int k=0; k < z_size_d; k++) {
          table_d[i][j][k].check_data();
        }
      }
    }
    //    cerr << "hash3 over ";
  }

};

#endif
