#include <iostream.h>

#include "obj_page.h"
#include "exc.h"
#include "obj_file.h"
#include "p_buffer.h"
#include "obj_stream.h"
#include "defs.h"

extern "C" {
  int bcmp();
};

///////////////////////
// CLASS ObjectPage //
/////////////////////

ObjectPage:: ObjectPage(const PageKey& p_key) {
  int state;
  PageNode* page_node = (object_stream->page_buffer()->get_cache(p_key, state));
  page_id = &(page_node->key());
  cache_d = &(page_node->cache());
  cache_d->lock();
  info_d = &(page_node->info());
  if (!state) {       // page needs to be retrieved from the file
    read_page();
  };
  UNEXPECTED ( (!cache_d && !info_d), 
	      "unexpected NULL value in class ObjectPage::ObjectPage()" )
}

ObjectPage:: ObjectPage(const PageKey& p_key, FirstNewPage fnp)
{
  PageNode* page_node = (object_stream->page_buffer()->new_cache(p_key));
  page_id = &(page_node->key());
  cache_d = &(page_node->cache());
  cache_d->lock();
  info_d = &(page_node->info());

  UNEXPECTED ( (!cache_d && !info_d), 
	      "unexpected NULL value in class ObjectPage::ObjectPage()" )
  make_new();
  cache_d->dirty(1);
}

//ObjectPage:: ObjectPage(const PageKey& p_key, NewPage np)
ObjectPage:: ObjectPage(NewPage np, ObjectPage& prev_page)
{

  page_num new_pnum = prev_page.file()->new_page_num(prev_page.pnum());
  PageKey p_key(prev_page.file(), new_pnum);

  PageNode* page_node = (object_stream->page_buffer()->new_cache(p_key));
  page_id = &(page_node->key());
  cache_d = &(page_node->cache());
  cache_d->lock();
  info_d = &(page_node->info());

  UNEXPECTED ( (!cache_d && !info_d), 
	      "unexpected NULL value in class ObjectPage::ObjectPage()" )
  make_new();

  ObjPageInfo* info = (ObjPageInfo*)info_d;

//   ObjPageInfo* p_info = (ObjPageInfo*)prev_page.info_d;
//   if (p_info->next_page) {
//     PageKey next_pid(p_key.file(), p_info->next_page);
//     ObjectPage next_page(next_pid);
//     ObjPageInfo* n_info = (ObjPageInfo*)next_page.info_d;

//     info->next_page=next_page.pnum();
//     n_info->prev_page=this->pnum();
//     next_page.cache_d->dirty(1);
//   }
//   else {
//     info->next_page=0;
//   }
//   info->prev_page=prev_page.pnum();
//   p_info->next_page=this->pnum();
//   prev_page.cache_d->dirty(1);

  cache_d->dirty(1);

}

ObjectPage:: ~ObjectPage() {

  cache_d->un_lock();

}

const char* ObjectPage:: operator[](slot_num snum)const{
  UNEXPECTED( (snum >= ((ObjPageInfo*)info_d)->free_slot),
	     "unexpected slot value in ObjectPage::operator[]" );
  s_2byte sl = SLOT_LOC(snum);
  s_2byte obj_loc;
  Cache& cache = *cache_d;
  pt2s_2byte(obj_loc, cache[sl]);
  return cache[obj_loc];
}

// check if enough space for the object of bytes size 
// and two bytes slot for it.
int ObjectPage:: has_space(o_2byte bytes) {
  ObjPageInfo* info = (ObjPageInfo*)info_d; // type_cast
  //  return ( bytes <= info->actual_free_bytes );
  return ( (bytes+SLOT_SIZE) <= info->actual_free_bytes );
}

o_2byte ObjectPage:: free_space() {
  ObjPageInfo* info = (ObjPageInfo*)info_d;
  return info->actual_free_bytes;
}

page_num ObjectPage:: last_obj_page() {
  return page_id->file()->last_obj_page(); }

int ObjectPage:: first_object() {
  return page_id->file()->first_obj_page();
}

int ObjectPage:: num_of_objects()const {

  ObjPageInfo* info = (ObjPageInfo*)info_d; // type_cast
  return (info->free_slot - 1);

}

void ObjectPage:: ptrs_lock(int y) { cache_d->ptrs_lock(y); }

int ObjectPage:: num_of_ptr_locks() { cache_d->num_of_ptr_locks(); }

o_2byte ObjectPage:: object_size(slot_num snum)const {
  Cache& cache = *cache_d;
  s_2byte sl = SLOT_LOC(snum);
  s_2byte obj_loc;
  pt2s_2byte(obj_loc, cache[sl]);
  char* loc = cache[obj_loc];
  o_2byte obj_size = object_size(loc);
  return obj_size;
}

o_2byte ObjectPage:: object_size(const char* d)const {
  o_2byte obj_size;
  pt2o_2byte(obj_size, OBJ_SIZE(d));
  return obj_size;
}

void op_temp() {}

slot_num ObjectPage:: create_object(o_2byte obj_size, const char* dat)const
{
  s_2byte space_needed = obj_size+SLOT_SIZE;
  UNEXPECTED( (object_size(dat) != obj_size),
	     "unexpected error (size mismatch) in ObjectPage:: create_object" );

  ObjPageInfo* info = (ObjPageInfo*)info_d; // type_cast 

  UNEXPECTED( (space_needed > info->actual_free_bytes),
             "unexpected error (no space left) in ObjectPage:: create_object" );

  //  cerr << "**** 1\n"; file()->usage(cerr);
  if ( space_needed > info->free_bytes 
       && space_needed <= info->actual_free_bytes ) {
    compress_page();
    //    cerr << "**** 2\n"; file()->usage(cerr);
    UNEXPECTED ( ( space_needed > info->free_bytes ), 
		 "unexpected error (while compressing) in ObjectPage:: create_object" );
  }
  
  slot_num snum = info->free_slot;
  Cache& cache = *cache_d;

  cache(info->free_loc, dat, obj_size);
  s_2byte sl = SLOT_LOC(info->free_slot);
  //  cerr << "**** 2.5\n"; file()->usage(cerr);
  cp_s_2byte_to_pt(cache[sl], info->free_loc);
  info->free_slot++;
  info->free_loc += obj_size;
  info->free_bytes -= space_needed;
  info->actual_free_bytes -= space_needed;
  //  cerr << "**** 300\n"; file()->usage(cerr);
  cache_d->dirty(1);

  //  cerr << "free_bytes=" << info->free_bytes << "\n\n\n";

  return snum;
}

slot_num ObjectPage:: alloc_space(o_2byte obj_size)const
{
  ObjPageInfo* info = (ObjPageInfo*)info_d; // type_cast 
  s_2byte space_needed = obj_size+SLOT_SIZE;
  
  // cerr << "**** 4\n"; file()->usage(cerr);
  if ( space_needed > info->free_bytes && space_needed <= info->actual_free_bytes ) {
    compress_page();
    // cerr << "**** 5\n"; file()->usage(cerr);
    UNEXPECTED ( ( space_needed > info->free_bytes ), 
		 "unexpected error in ObjectPage:: alloc_space 1" );
  }
  
  slot_num snum = info->free_slot;

  Cache& cache = *cache_d;

  o_2byte zero_size=0;
  cp_o_2byte_to_pt(cache[info->free_loc], obj_size);      

  cp_s_2byte_to_pt(cache[SLOT_LOC(info->free_slot)], info->free_loc);
 
  info->free_slot++;
  info->free_loc += obj_size;

  info->free_bytes -= space_needed;
  info->actual_free_bytes -= space_needed;

  // cerr << "**** 6\n"; file()->usage(cerr);
  cache_d->dirty(1);

  return snum;
}


void ObjectPage:: update_object(slot_num snum, o_2byte obj_size,
				const char* d) {
  UNEXPECTED( (object_size(d) != obj_size),
	     "unexpected error (size mismatch) in ObjectPage:: update_object" )
  Cache& cache = *cache_d;
  s_2byte sl = SLOT_LOC(snum);
  s_2byte obj_loc;
  pt2s_2byte(obj_loc, cache[sl]);
	
  UNEXPECTED( (bcmp(OBJ_SIZE(d), OBJ_SIZE(cache[obj_loc]), OI_SIZE)),
	     "updating a different size object in ObjectPage::update_object" );
  // cerr << "**** 7\n"; file()->usage(cerr);
  cache(obj_loc, d, obj_size);
  // cerr << "**** 8\n"; file()->usage(cerr);
  cache_d->dirty(1);
}

slot_num ObjectPage:: realloc_space(slot_num snum, o_2byte obj_size) {

  ObjPageInfo* info = (ObjPageInfo*)info_d; // type_cast 
  Cache& cache = *cache_d;
  s_2byte space_needed = obj_size+SLOT_SIZE;
  o_2byte size = object_size(snum);

  // cerr << "**** 9\n"; file()->usage(cerr);
  if (obj_size <= size) {
    s_2byte obj_loc;
    pt2s_2byte(obj_loc, cache[SLOT_LOC(snum)]);
    cp_o_2byte_to_pt(cache[obj_loc], obj_size); 
    info->actual_free_bytes += (size-obj_size);
  }
  else {
    o_2byte po_size = POINTER_OBJECT_SIZE;	
    o_2byte req_space = (obj_size+po_size-size);
    if (has_space(req_space)) {
      s_2byte obj_loc;
      pt2s_2byte(obj_loc, cache[SLOT_LOC(snum)]);
      cp_o_2byte_to_pt(cache[obj_loc], po_size); 
      info->actual_free_bytes += (size-po_size);
      // cerr << "**** 10\n"; file()->usage(cerr);
      slot_num snum_f = alloc_space(obj_size);
      modify_object(snum, pnum(), snum_f, F_TYPE);
      return snum_f;
    }
    //    cerr << "**** 10.2\n"; file()->usage(cerr);
    return 0;
  }
  // cerr << "**** 10.5\n"; file()->usage(cerr);
  cache_d->dirty(1);

  return snum;
}

int ObjectPage:: modify_object(slot_num snum, o_2byte obj_size,
			       const char* d) {
  int success=1;

  UNEXPECTED( (object_size(d) != obj_size),
	     "unexpected error (size mismatch) in ObjectPage:: modify_object" );

  // cerr << "**** 11\n"; file()->usage(cerr);
  ObjPageInfo* info = (ObjPageInfo*)info_d; // type_cast 
  Cache& cache = *cache_d;

  o_2byte size = object_size(snum);

  if (obj_size <= size) {
    s_2byte obj_loc;
    pt2s_2byte(obj_loc, cache[SLOT_LOC(snum)]);
    cache(obj_loc, d, obj_size);
    info->actual_free_bytes += (size-obj_size);
    // cerr << "**** 11.5\n"; file()->usage(cerr);
  }
  else {
    o_2byte po_size = POINTER_OBJECT_SIZE;	
    o_2byte req_space = (obj_size+po_size-size);
    if (has_space(req_space)) {
      s_2byte obj_loc;
      pt2s_2byte(obj_loc, cache[SLOT_LOC(snum)]);
      cp_o_2byte_to_pt(cache[obj_loc], po_size); 
      info->actual_free_bytes += (size-po_size);
      // cerr << "**** 11.6\n"; file()->usage(cerr);
      slot_num snum_f = create_object(obj_size, d);
      modify_object(snum, pnum(), snum_f, F_TYPE);
    }
    else success = 0;
  }
  // cerr << "**** 12\n"; file()->usage(cerr);
  cache_d->dirty(1);

  return success;
}

void ObjectPage:: modify_object(slot_num snum, 
				page_num pnum_p, 
				slot_num snum_p,
				u_1byte typ) {
  ObjPageInfo* info = (ObjPageInfo*)info_d; // type_cast 
  Cache& cache = *cache_d;

  // cerr << "**** 13.77\n"; file()->usage(cerr);

  s_2byte sl = SLOT_LOC(snum);
  s_2byte obj_loc;
  pt2s_2byte(obj_loc, cache[sl]);
  char* loc = cache[obj_loc];

  o_2byte size = object_size(loc);

  UNEXPECTED( (size < POINTER_OBJECT_SIZE),
	     "unexpected error in ObjectPage:: modify_object 3" );

  // cerr << "**** 13\n"; file()->usage(cerr);
  o_2byte po_size = POINTER_OBJECT_SIZE;
  info->actual_free_bytes += (size-po_size);
  cp_o_2byte_to_pt(loc, po_size);      
  // cerr << "**** 14\n"; file()->usage(cerr);
  modify_object(loc, pnum_p, snum_p, typ);
  cache_d->dirty(1);

}

void ObjectPage:: modify_object(char* loc,  
				page_num pnum_p, 
				slot_num snum_p,
				u_1byte typ) { // private function

  // cerr << "**** 21\n"; file()->usage(cerr);

  UNEXPECTED( (object_size(loc) != POINTER_OBJECT_SIZE),
	     "unexpected error in ObjectPage:: modify_object 5" );
  loc += OI_SIZE;
  *(u_1byte*)loc=typ;
  loc += TI_SIZE;
  cp_p_4byte_to_pt(loc, pnum_p);
  loc += 4;
  cp_s_2byte_to_pt(loc, snum_p);

  // cerr << "**** 22\n"; file()->usage(cerr);
}

void ObjectPage:: compress_page(void)const{

#ifdef CHECK_UNEXPECTED
  if (cache_d->is_ptrs_locked()) {
    dberror << "unexpected error in ObjectPage:: do not compress page, as pointers to the page are in use";
    throw_error();
  }
#endif
  // cerr << "**** 15\n"; file()->usage(cerr);
  // cerr << "*** compressing " << pnum() << "\n";
  s_2byte curr_ptr=FIRST_OBJ_LOC_IN_OBJ_PAGE;
  s_2byte i;

  ObjPageInfo* info = (ObjPageInfo*)info_d; // type_cast

  s_2byte sl = SLOT_LOC(info->free_slot);
  Cache& cache = *cache_d;
	
  for(i=SLOT_LOC(1); i > sl; i -= SLOT_SIZE) {

    s_2byte obj_loc;
    pt2s_2byte(obj_loc, cache[i]);
    char* buf = cache[obj_loc];
    o_2byte obj_size = object_size(buf);
 
    if ( curr_ptr == obj_loc ) {
      curr_ptr += obj_size;
    }
    else if ( curr_ptr < obj_loc ) {
      cache(curr_ptr, obj_loc, obj_size);
      cp_s_2byte_to_pt(cache[i], curr_ptr);

//pt2s_2byte(obj_loc, cache[i]);

      curr_ptr += obj_size;
    }
  };
  sl += SLOT_SIZE;
  int space_left = sl - curr_ptr;

  //  if (info->free_loc < curr_ptr || 
  //      info->actual_free_bytes != (sl - curr_ptr)) op_temp();

  UNEXPECTED( ( space_left < 0 ),
	      "unexpected page overflow in ObjectPage:: compress_page");
  UNEXPECTED( (info->free_loc < curr_ptr || 
  	       info->actual_free_bytes != (sl - curr_ptr)),
  	      "unexpected error (1) in ObjectPage:: compress_page");

  // cerr << "free_loc: " << info->free_loc 
  //       << " free_bytes: " << info->actual_free_bytes
  //         << " curr_ptr: " << curr_ptr 
  //         << " sl: " << sl  << " diff: " << sl-curr_ptr << "\n";

  info->free_loc = curr_ptr;
  info->free_bytes = sl-curr_ptr;
  info->actual_free_bytes = info->free_bytes; 
  cache_d->dirty(1);
}

void ObjectPage:: usage(ostream& os)const{
  os << "id " << page_id->pnum() << " => ";
  os << ", Objects: " << num_of_objects() << " ";
  info_d->usage(os);
}

void ObjectPage:: output_data(ostream& os)const{
  os << "Page Id: " << page_id->pnum() << " => ";
  os << *info_d;
}

int ObjectPage:: page_ok()const{
  ObjPageInfo* info = (ObjPageInfo*)info_d; // type_cast
  s_2byte sl = SLOT_LOC(info->free_slot);
  Cache& cache = *cache_d;	
  s_2byte used_space = OBJ_PAGE_HDR_SIZE + (info->free_slot-1)*SLOT_SIZE;
  int i;
  for(i=SLOT_LOC(1); i > sl; i -= SLOT_SIZE) {
    s_2byte obj_loc;
    //    cerr << "i: " << i << " ";
    pt2s_2byte(obj_loc, cache[i]);
    char* buf = cache[obj_loc];
    o_2byte obj_size = object_size(buf);
    used_space += obj_size;
    //cerr << "obj_loc: " << obj_loc << "\n";
  };
  //  cerr << "page_ok\n";
  int space_left = PAGE_SIZE-used_space;
  if (info->actual_free_bytes != space_left) {
    cerr << " free_bytes: " << info->actual_free_bytes
	 << " un_used_space: " << PAGE_SIZE-used_space<< "\n";
    op_temp();  
    //    throw_error();
    return 0;
  }
  return 1;
}
