
#ifndef obj_page_hdr
#define obj_page_hdr

class ObjectPage;

#include "page.h"

struct FirstNewPage;
class PageKey;

class NewPage {};


///////////////////////
// CLASS ObjectPage //
/////////////////////
class ObjectPage : public Page {

public:
  ObjectPage(const PageKey& p_key);
  ObjectPage(const PageKey&, FirstNewPage);
  ObjectPage(NewPage np, ObjectPage& prev_page);
	       // create new page and insert it after prev_page.
  ~ObjectPage();

  int page_type()const{ return ObjPageInfo::type_d; }

  const char* operator[](slot_num snum) const;

  int first_object();
  int num_of_objects()const;
  void ptrs_lock(int y);
  int num_of_ptr_locks();

  int has_space(o_2byte bytes);
  o_2byte free_space();
  page_num last_obj_page();

  slot_num create_object(o_2byte obj_size, const char* data)const;
  slot_num alloc_space(o_2byte obj_size)const;
  void update_object(slot_num snum, o_2byte obj_size,
		     const char* data);
       // update_object is used when the modified object size is
       // same as before.

  slot_num realloc_space(slot_num snum, o_2byte obj_size);
      // if the object can be relocated in the same page, 
      // at the same location or as a forwarded object 
      // it returns snum where it was relocated  on success.
      // if no space left on the page it returns 0 and the
      // object is left untouched.
  int modify_object(slot_num snum, o_2byte obj_size,
		     const char* d);
      // if the object can be modified in the same page, 
      // at the same location or as a forwarded object 
      // it returns 1 on success.
      // if no space left on the page it returns 0 and the
      // object is left untouched.
  void modify_object(slot_num snum, 
		     page_num pnum_p, 
		     slot_num snum_p,
		     u_1byte typ); // typ = F_TYPE or OID_TYPE
      // modifies the object snum as forwarded object, or as a
      // pointer to another object.
  void usage(ostream& os)const;

protected:
  void modify_object(char*, page_num pnum_p, 
		     slot_num snum_p,
		     u_1byte typ); // typ = F_TYPE or OID_TYPE

  void compress_page(void)const;
  o_2byte object_size(slot_num snum)const;
  o_2byte object_size(const char* d)const;

  void output_data(ostream& os)const;

public:
  int page_ok()const;
};

#endif
