#include "obj_stream.h"
#include "o_buffer.h"
#include "obj_file.h"
#include "p_buffer.h"

extern int save_top_only;

FileNode:: ~FileNode() {
}

extern jmp_buf env;

//////////////////////
// CLASS ObjStream //
////////////////////
int ObjStream:: page_hash_table_size[2] = {3,13}; 
	      // [0] = file_hash_size, [1] = page_hash_size
int ObjStream:: obj_hash_table_size[3] = {3,13,7};
	    // [0] = file_hash_size, [1] = page_hash_size, [3] = object_hash_size

static int NUMOFNODES(REAL i, int size, int mem_size) {
  i = (i>90)?90:(i<0)?0:i;
  //  return (int)floor((alloc_words*((float)i/100))/size);
  return (int)floor((mem_size*((float)i/100))/size);
}

ObjStream:: ObjStream(REAL max_pages, REAL obj_y_hash_size)	   
     //	   :curr_obj_file(NULL)						   
{
  long mem_size = sizeof(mem_base)*(mem_limit-mem_base);
  int max_num_of_pages = NUMOFNODES(max_pages, sizeof(PageNode), mem_size);
  //  cerr << max_num_of_pages << "\n";
  max_num_of_pages = (max_num_of_pages<3)?3:max_num_of_pages;
  //  if ((max_num_of_pages*sizeof(PageNode)) > alloc_words) {
  if ((max_num_of_pages*sizeof(PageNode)) > mem_size) {
    Errorline("memory size too small for database operations -- get a bigger Life!\n");
    abort_life(TRUE);
    longjmp(env,FALSE); 
  };
  //  cerr << "max number of pages: " << max_num_of_pages << "\n";
  int page_hash_size = (100*max_num_of_pages)/100;
  //  cerr << "hash: " << page_hash_size << "\n";
  page_hash_table_size[1] = (page_hash_size<13)?13:page_hash_size;
  page_buffer_d = new PageBuffer(max_num_of_pages,
				 page_hash_table_size[1],
				 page_hash_table_size[0]);
  //  cerr << "2\n";
  int obj_z_size = (int)floor((OBJ_SIZE_MAX/POINTER_OBJECT_SIZE)/32);
  obj_hash_table_size[2] = (obj_z_size < 7)?7:obj_z_size;

  //  int obj_y_size = (int)floor( (alloc_words*(float)obj_y_hash_size/100)/
  //    (sizeof(ObjectNode)+sizeof(psi_term)));
  obj_y_hash_size = (obj_y_hash_size>90)?90:
    (obj_y_hash_size<0)?0:obj_y_hash_size;
  //  cerr << obj_y_hash_size << "\n";
  int obj_y_size = (int)floor( (mem_size*(float)obj_y_hash_size/100)/
    (sizeof(ObjectNode)+sizeof(psi_term)));
  obj_hash_table_size[1] = (obj_y_size<13)?13:obj_y_size;
  //  cerr << "obj: " << obj_z_size << " " << obj_y_size << "\n";
  object_buffer_d = new ObjectBuffer(obj_hash_table_size[2],
				     obj_hash_table_size[1],
				     obj_hash_table_size[0]);
}

ObjStream:: ~ObjStream() {
  //  cerr << "obj_stream destructor called" << "\n";
  flush_obj_stream(1);
  UNEXPECTED( (!object_buffer()->is_flushed()),
	      "unexpected error (could not save persistent terms) in ObjStream:: ~ObjStream()");
}

int ObjStream:: flush_obj_stream(int save_all) {
  int flushed=1;
//   FileNode* fn=(FileNode*)file_table.first();
//   while(fn) {
//     flushed=fn->file()->flush_dbpers_vars(save_all);
//     fn=(FileNode*)fn->next1();
//   }

  flushed=object_buffer()->flush_object_buffer(save_all);
  page_buffer()->flush_pages();
  
  FileNode* fn=(FileNode*)file_table.first();
  while(fn) {
    fn->file()->flush_file_hdr();
    fn=(FileNode*)fn->next1();
  };
  return flushed;
}

void ObjStream:: check_data() {

  object_buffer()->check_data();
  page_buffer()->check_data();

//   FileNode* fn=(FileNode*)file_table.first();
//   while(fn) {
//     fn->file()->check_dbpers_vars();
//     fn=(FileNode*)fn->next1();
//   }

}

extern long odb_verbose;
int ObjStream:: evict_objects(int min_pages, int min_objects) {

  //  cerr << "min number of pages: " << min_num_of_pages << "\n";
  //  cerr << "min number of objects: " << min_num_of_objects << "\n";

  if (odb_verbose) {
    cerr << "*** Evicting Pesistent Terms\n";
    cerr << "*** Begin ";
    output_heap_info(cerr);
    cerr << "\n";
  }

  int objects_evicted = object_buffer()->evict_objects(min_objects);
  int pages_evicted = page_buffer()->evict_pages(min_pages);

  if (odb_verbose) {
    cerr << "*** End ";
    output_heap_info(cerr);
    cerr << "\n";
  }
  return (objects_evicted || pages_evicted);
}

// outputted on user calling odb_verbose.
void ObjStream:: output_heap_info(ostream& os) {
    os << "[page buffer=> " << page_buffer()->size()/1000.0
      << "K (Pages:" << page_buffer()->length() << "), ";
    os << "object buffer=> " << object_buffer()->size()/1000.0
      << "K (Objects:" << object_buffer()->length() << ")]";    
}

extern int fetch_depth;

void ObjStream:: output_statistics(ostream& os) {
    os << "Page buffer: " 
       << "[" << page_hash_table_size[1] << " x "
       << page_hash_table_size[0] << "] "
       << page_buffer()->size()/1000.0
      << "K (Pages: " << page_buffer()->length() 
	<<  ", Max: " << page_buffer()->max_size()/1000.0 << "K"
       << ", Max pages: " << page_buffer()->max_length() << "), ";
    os << "\n";
    os << "Object buffer: " 
       << "[" << obj_hash_table_size[2] << " x "
       << obj_hash_table_size[1] << " x "
       << obj_hash_table_size[0] << "] "
       << object_buffer()->size()/1000.0
      << "K (Objects: " << object_buffer()->length() 
	<< ", Fetch depth: " << fetch_depth << ")";    
}

ObjFile* ObjStream:: open_file(char* str) {
//   curr_obj_file = new ObjFile(str);
//   file_table.append(new FileNode(curr_obj_file));
//   return curr_obj_file;
  ObjFile* obj_file = new ObjFile(str);
  file_table.append(new FileNode(obj_file));
  return obj_file;
}

int ObjStream:: close_file(ObjFile* file) {

  int objects_evicted = object_buffer()->evict_objects(file);
  page_buffer()->evict_pages(file);
  if (objects_evicted) {
    //    if (curr_obj_file==file) 
    //      curr_obj_file=NULL;
    file->close();
  }
  return objects_evicted;
}

void ObjStream:: usage(ostream& os)const {
  FileNode* fn=(FileNode*)file_table.first();
  os << "*** page hash size: [" << page_hash_table_size[0]
    << ", " << page_hash_table_size[1] << "]\n";
  os << "*** object hash size: [" << obj_hash_table_size[0]
     << ", " << obj_hash_table_size[1] 
    << ", " << obj_hash_table_size[2] << "]\n";
 
  os << "*** page buffer: "; page_buffer_d->usage(os); os << "\n";
  os << "*** object buffer: "; object_buffer_d->usage(os); os << "\n";
  
  while(fn) {
    os << "*** File: ";
    fn->file()->usage(os);
    os << "\n\n";
    fn=(FileNode*)fn->next1();
  };
  os << "\n\n";
}

void ObjStream:: output_data(ostream& os) const {
  usage(os);
}

ostream& operator<<( ostream & os, const ObjStream& n ) {
  n.output_data(os);
  return os;
}






