#ifndef odb_hdr
#define odb_hdr

#include <iostream.h>

#include "extern-c++.h"
#include "db.h"

extern int byte_order_ok;

extern "C" {
  void bcopy();
};

#define bcopy_4(s, d) \
{ \
   const char* src = s; char* dest = d; \
   if (byte_order_ok) \
       bcopy(src, dest, 4); \
   else \
       {dest[0]=src[3]; dest[1]=src[2]; dest[2]=src[1]; dest[3]=src[0];} }

#define bcopy_2(s, d) \
{ \
        const char* src = s; char* dest = d; \
        if (byte_order_ok) \
	    bcopy(src, dest, 2); \
	else \
            {dest[0]=src[1]; dest[1]=src[0];} } 

///////////////////////////////////////////////////////////////////////

#define u_1byte unsigned char

#define u_2byte unsigned short
#define pt2u_2byte(v, cp) bcopy_2(cp, (char*)&v)
#define cp_u_2byte_to_pt(cp, v) bcopy_2((char*)&v, cp)

#define u_4byte unsigned int
#define pt2u_4byte(v, cp) bcopy_4(cp, (char*)&v)
#define cp_u_4byte_to_pt(cp,v) bcopy_4((char*)&v, cp)

#define o_2byte unsigned short
#define pt2o_2byte(v, cp) bcopy_2(cp, (char*)&v)
#define cp_o_2byte_to_pt(cp, v) bcopy_2((char*)&v, cp)

#define s_2byte unsigned short
#define pt2s_2byte(v, cp) bcopy_2(cp, (char*)&v)
#define cp_s_2byte_to_pt(cp, v) bcopy_2((char*)&v, cp)

#define p_4byte unsigned int
#define pt2p_4byte(v, cp) bcopy_4(cp, (char*)&v)
#define cp_p_4byte_to_pt(cp,v) bcopy_4((char*)&v, cp)

#define v_4byte unsigned int
#define pt2v_4byte(v, cp) bcopy_4(cp, (char*)&v)
#define cp_v_4byte_to_pt(cp,v) bcopy_4((char*)&v, cp)

/////////////////////////////////////////////////////////////////////////

typedef v_4byte vol_num;
typedef p_4byte page_num;
typedef s_2byte slot_num;

class LogType {};

////////////////////////
// STRUCT TwoByteLog //
//////////////////////
struct TwoByteLog {
private:
  const u_1byte log_value;
  const u_2byte value;
public:
  TwoByteLog(const u_1byte& v, LogType lt):log_value(v),value(1<<v){}
  // The second argument of the constructor is ignored. It is there
  // to prevent constructor being used as a implicit converter function
  // as it might happen in case of friend functions below.
  operator u_2byte() const { return value; }
  unsigned long operator*(unsigned long v)const 
    { return v << log_value; }
  friend unsigned long operator*(unsigned long v, const TwoByteLog& v2)
    {return v<<v2.log_value;}
  unsigned long operator/(unsigned long v)const 
    { return v >> log_value; }
  friend unsigned long operator/(unsigned long v, 
				 const TwoByteLog& v2) 
    {return v>>v2.log_value;}
};


////////////////////////
// STRUCT OneByteLog //
//////////////////////
struct OneByteLog {
private:
  const u_1byte log_value;
  const u_1byte value;
public:
  OneByteLog(const u_1byte& v, LogType lt):log_value(v),value(1<<v){}
  // The second argument of the constructor is ignored. It is there
  // to prevent constructor being used as a implicit converter function
  // as it might happen in case of friend functions below.
  operator u_1byte() const { return value; }
  unsigned long operator*(unsigned long v)const 
    {  return v << log_value; }
  friend unsigned long operator*(unsigned long v, 
				 const OneByteLog& v2) 
    {return v<<v2.log_value;}
  unsigned long operator/(unsigned long v)const 
    { return v >> log_value; }
  friend unsigned long operator/(unsigned long v, 
				 const OneByteLog& v2) 
   {return v>>v2.log_value;}
};


////////////////////////
// DbFileConstants ////
//////////////////////

enum {DB_FILE=0, OBJ_FILE=1};
enum {PAGE=0, OBJ_PAGE=1};

//-------------- PAGE LAYOUT -----------------//
//const TwoByteLog PAGE_SIZE(10, LogType());
//const u_2byte PAGE_SIZE=1024;
const u_2byte PAGE_SIZE=8192;
//const u_2byte PAGE_SIZE=256;
  // PAGE_SIZE is 2 power k.

const u_1byte OBJ_PAGE_HDR_SIZE=17;
//const OneByteLog SLOT_SIZE(1, LogType());
const u_1byte SLOT_SIZE=2;
           // byte 0 = page_type.
	   // byte 1,2,3,4 next_page.
	   // byte 5,6,7,8 prev_page.
	   // byte 9,10 = free_slot on obj_page.
	   // byte 11,12 = free_loc to place the object on the page.
	   // byte 13,14 = free_bytes available after
	   // the last object on the page to place the object.
	   // byte 15,16 = actual_free_bytes available on the page
	   // to place the object.
	   // objects are placed after the OBJ_PAGE_HDR_SIZE bytes, while the 
	   // slots starting from end of the page. Each slot of SLOT_SIZE  
	   // consists of slot_num of two-bytes.

#define PAGE_TYPE(page) ((u_1byte*)page)[0]
#define SLOT_LOC(S) (PAGE_SIZE-(S*SLOT_SIZE))
#define FIRST_OBJ_LOC_IN_OBJ_PAGE OBJ_PAGE_HDR_SIZE
//----------------------------------------//

//------------- OBJECT LAYOUT ---------------//

const o_2byte  OBJ_SIZE_MAX =  PAGE_SIZE-OBJ_PAGE_HDR_SIZE-SLOT_SIZE; 
	     // Objects larger than this don not fit into a page.

//*** OBJ_SIZE *****************************//
const u_1byte OI_SIZE=2;
          // byte 0,1 = object_size.
// if object_size < POINTER_OBJECT_SIZE the object_size is set
// to POINTER_OBJECT_SIZE. 


//*** Type( 1 byte size ) ***************//
const u_1byte TI_SIZE=1;
enum {OBJECT_TYPE_MASK=128}; // 128 = 10000000
        // 1=large_object, 0=small_object
enum {PSI_MASK=64}; // 64 = 01000000
// if 7th bit is 1 Type is followed by sort_size(SI), else by the sort
// and size of root_sort is given by OBJ_SIZE-TI_SIZE-OI_SIZE-PADDED_SIZE.
enum {SORT_TYPE_MASK=56}; // 56 = **111***
enum {NONE_TYPE=0, SORT_TYPE=8, STRING_TYPE=16, REAL_TYPE=24};
// 0 = **000***, 8 = **001***, 16 = **010***, 24 = **011***
enum {OID_TYPE=32, F_TYPE=40};
// 32 = **100***, 40 = **101***

// F_TYPE indicates forwarded object
// OID_TYPE indicates it is a pointer to another object
enum {PADDED_MASK=7}; // 7 = *****111
// the lower three bits are used to for padded_size value, 
// since if object_size < POINTER_OBJECT_SIZE the object_size is set
// to POINTER_OBJECT_SIZE.
const u_1byte PADDED_SIZE_MAX=7;

//*** sort_size(SI) *******************//
const u_1byte SI_SIZE=2;
const u_2byte SI_MAX =65535; //65535 = 11111111 11111111

//*** Number of Fields *******************//
const u_1byte FNUM_SIZE = 2;
const u_2byte FNUM_MAX=65535;

//*** Field_Index ************************//
//const OneByteLog FI_SIZE(1, LogType());
const u_1byte FI_SIZE = 2;
const u_2byte FI_MAX=32767; // 32767 = *1111111 11111111
enum {FI_LOOP_MASK=128}; // 128 = 1*******
// if 8th bit is one then the field value is a loop 
// to the same object.
enum {FI_SIZE_MASK=127}; // 127 = *1111111

//*** OID_POINTER *********************//
const u_1byte OID_SIZE = 6; 
// The size of this pointer is 6 bytes
// page_num(4 byte) slot_num(2 byte).

#define POINTER_OBJECT_SIZE (OI_SIZE+TI_SIZE+OID_SIZE) //=9

          // ---------------
	  // |OBJ_SIZE(OI) |
	  // ---------------

	  // ------------------------------------
	  // | Type | sort_size(SI) | root sort | 
	  // ------------------------------------
	  // ----------------------------------------------
	  // # of fields(FNUM) | array of field_index(FI) |
	  // ----------------------------------------------
	  // ------------------------
	  //  OID_POINTER | field\0 |
	  // ------------------------
// if object_size < POINTER_OBJECT_SIZE

#define OBJ_SIZE(C) C

//----------------------------------------//

//////////////////////
// STRUCT PageInfo //
////////////////////
struct PageInfo {

  static const u_1byte type_d;

  PageInfo();

  virtual void init(char* buf) = 0;
  virtual int read_hdr(char* buf);
  virtual void write_hdr(char* buf)const;
  virtual int type()const { return type_d; }

  virtual void  usage(ostream& os)const=0;
  virtual void output_data(ostream&)const=0;
  friend ostream& operator<<(ostream&, const PageInfo&);
};

/////////////////////////
// STRUCT ObjPageInfo //
///////////////////////
struct ObjPageInfo:public PageInfo {

  static const u_1byte type_d;

  page_num next_page; // next_page of zero is used to indicate NULL
  page_num prev_page; // prev_page of zero is used to indicate NULL
  slot_num free_slot;

  s_2byte free_loc;
  s_2byte free_bytes;
  s_2byte actual_free_bytes;

  ObjPageInfo();

  void init(char* buf);
  int read_hdr(char* buf);
  void write_hdr(char* buf)const;
  int type()const { return type_d; }

  void  usage(ostream& os)const;
  void output_data(ostream&)const;
};

////////////////////////
// STRUCT DbFileInfo //
//////////////////////
struct DbFileInfo {
  static const u_1byte type_d;

  virtual void init()=0;
  virtual void read_hdr(const char* buf)=0;
  virtual void write_hdr(char* buf)const=0;
  virtual int type()const { return type_d; }

  friend ostream& operator<<(ostream&, const DbFileInfo&);

protected:
  virtual void output_data(ostream&)const=0;
};

////////////////////////
// STRUCT ObjFileInfo //
//////////////////////
struct ObjFileInfo:public DbFileInfo {

  static const u_1byte type_d;
  int state_d;
  static const page_num first_obj_page;
  page_num last_obj_page;
  page_num num_of_obj_pages;

  static const u_2byte FILE_HDR_SIZE;
  
  ObjFileInfo();

  static page_num bid2pnum(const p_4byte&);
  static p_4byte pnum2bid(const page_num&);

  void init();
  void write_hdr(char* buf)const;
  void read_hdr(const char* buf);
  int type()const { return type_d; }

  enum { DIRTY_MASK=1 };
  void modified(int a) {
    if (a) state_d |= DIRTY_MASK;
    else state_d &= ~DIRTY_MASK;
  }
  int is_modified() { return state_d&DIRTY_MASK; }

protected:
  void output_data(ostream&)const;

};

#endif
