#include "p_buffer.h"
#include "obj_file.h"

extern "C" {

void check_bytedata(GENERIC*);

}

#define BYTE_COPY(a, b, l) bcopy((const char*)(b), (char*)(a), l)

//////////////////
// CLASS Cache //
////////////////

void Cache:: operator()(s_2byte index, const char* d, s_2byte size) {

  UNEXPECTED( ((index + size) >= PAGE_SIZE),
	     "unexpected error in Cache:: operator(), cache access violation");
  BYTE_COPY(loc_d+index, d, size);
}

void Cache:: operator()(s_2byte to, s_2byte from, s_2byte size) {

  UNEXPECTED( ( (to + size) >= PAGE_SIZE || (from +size) >= PAGE_SIZE ),
	     "unexpected error in Cache:: operator(), cache access violation");
  BYTE_COPY(loc_d+to, loc_d+from, size);
}

void Cache:: lock() {
  lock_d++;
}

int Cache:: num_of_locks() {
  return lock_d;
}

void Cache:: un_lock() {
  if (lock_d > 0) lock_d--;
}

void Cache:: ptrs_lock(int y) {
  if(y) ptrs_lock_d++;
  else if (ptrs_lock_d > 0) ptrs_lock_d--;
}

int Cache:: num_of_ptr_locks() {
  return ptrs_lock_d;
}

/////////////////////
// CLASS PageNode //
///////////////////
GENERIC PageNode:: alloc_ptr=NULL;
int PageNode:: max_num=5;
int PageNode:: n_of_nodes_d=0;

PageNode:: ~PageNode() {
//  cerr << "PageNode:: ~PageNode()\n";
}

void PageNode:: flush_page() {
  if (key_d.file()->file_writable()) {
    if (cache_d.is_dirty()) {
      print_debug((cerr << "*** Flushing page of " << key_d << "\n"));
      info_d.write_hdr(cache_d);
      page_num pnum = key_d.pnum();
      ObjFile* f = key_d.file();
      
      f->seek_page(pnum);
      key_d.file()->write_file(cache_d, cache_d.size());
      cache_d.dirty(0);
    }
    else {
      print_debug((cerr << "*** " << key() << " is unchanged" << "\n"));
    }
  }
}

void PageNode:: output_data(ostream& os)const {
  os << " PageNode: ";
  B_FL_Node:: output_data(os);
  os << key_d;
}

///////////////////////
// CLASS PageBuffer //
/////////////////////

PageBuffer:: PageBuffer(int i,
			int ph_size,  int fh_size):
               table(new PageHashTable(fh_size, ph_size?ph_size:i))
{
  PageNode::max_num=i;
}

PageBuffer:: ~PageBuffer() {
}

PageNode* PageBuffer:: get_cache(const PageKey& p_key, int& state) {
  PageNode* page_node;

  state=0;
  if (page_node = table->fetch(p_key)) {
    mv2end(page_node);    
    state=1;
  }
  else {
    page_node =  new_cache(p_key);
  }
  return page_node;
}

PageNode* PageBuffer:: new_cache(const PageKey& p_key) {
  PageNode* page_node=NULL;
  if (length() >= max_length()) page_node=free_a();
  if (!page_node) {
    page_node = new PageNode(p_key);
    // cerr << "*** number of page nodes: " << length() << "\n";
    UNEXPECTED( (length() > (max_length)+1),
		"in PageBuffer:: new_cache(const PageKey& p_key), number of page nodes exceeded" );
  }
  else {
    UNEXPECTED( (table->fetch(page_node->key())),
		"unexpected error in PageBuffer:: new_cache(const PageKey& p_key)");
    (*page_node)(p_key);
  }
  table->insert(page_node);
  append(page_node);
  return page_node;
}

PageNode* PageBuffer:: get_first_unlocked_node() {
  PageNode* page_node = first_page();
  while(page_node) {
    if ( !(page_node->cache().is_locked()) ) {
      return page_node;
    }
    page_node = next_page(page_node);
  }
  return NULL;
}

PageNode* PageBuffer:: get_last_unlocked_node() {
  PageNode * page_node = (PageNode*) last();
  while(page_node) {
    if ( !(page_node->cache().is_locked()) ) {
      return page_node;
    }
    page_node = prev_page(page_node);
  }
  return NULL;
}

PageNode* PageBuffer:: free_a() {
  PageNode * page_node = get_first_unlocked_node();
  if (page_node) {
    page_node->flush_page();
    page_node->remove();
  }
  return page_node;
}

PageNode* PageBuffer:: free_e() {
  PageNode * page_node = get_last_unlocked_node();
  if (page_node) {
    page_node->flush_page();
    page_node->remove();
  }
  return page_node;
}

void PageBuffer:: flush_pages() {
  PageNode* page_node = first_page();
  while(page_node) {
    page_node->flush_page();
    page_node=next_page(page_node);
  }
}

long p_evict_count=0;
long page_evict_count=0;

int PageBuffer:: evict_pages(int min_num) {
  UNEXPECTED( (FList:: length() != length()),
	      "unexpected error in PageBuffer:: evict_pages()" );

  int count = (min_num * length())/100;
  int count_val=count;
  if (count) {
    p_evict_count++;
    PageNode* next_node=NULL;
    PageNode* page_node = first_page();
    while(page_node && count > 0) {
      count--;
      page_node->flush_page();
      next_node=next_page(page_node);
      delete page_node;
      page_node=next_node;
    }
    
    int pages_evicted = count_val-count;
    page_evict_count += pages_evicted;
    UNEXPECTED( (FList:: length() != length()),
		"unexpected error in PageBuffer:: evict_pages()" );
    return (pages_evicted);
  }
  return 0;
}

void PageBuffer:: evict_pages(ObjFile* obj_file) {
  UNEXPECTED( (FList:: length() != length()),
	      "unexpected error in PageBuffer:: evict_pages()" );

  PageNode* next_node=NULL;
  PageNode* page_node = first_page();
  while(page_node) {
    next_node=next_page(page_node);
    if (page_node->key().file() == obj_file) {
      page_node->flush_page();
      delete page_node;
    }
    page_node=next_node;
  }

  UNEXPECTED( (FList:: length() != length()),
	      "unexpected error in PageBuffer:: evict_pages()" );
}

void PageBuffer:: check_data() {
  //  cerr << "pagebuffer ";
  table->check_data();
  FList:: check_data();
  //  cerr << "pagebuffer over ";
  PageNode* pn = first_page();
  while(pn) {
    pn->check_data();
    pn = next_page(pn);
  }
}

// int PageBuffer:: buf_length() {
//   int c=0;
//   PageNode* obn = first_page();
//   while(obn) {
//     c++;
//     obn = next_page(obn);
//   }
//   return c;
// }

void PageBuffer:: usage(ostream& os) {
  os << "number of pages in buffer is " << length() << 
    "(" << max_length() << ")" << "\n";
  table->usage(os);
}
