#ifndef psi_term_hdr
#define psi_term_hdr

#include "extern-c++.h"
extern "C" {

#include "Source/modules.h"
#include "Source/trees.h"
#include "Source/login.h"
#include "Source/lefun.h"
#include "Source/token.h"
#include "Source/memory.h"
#include "Source/built_ins.h"

}

#define BYTEDATA_SIZE(B) *(unsigned long*)B
#define BYTEDATA_DATA(B) ((char*)B + sizeof(unsigned long))
#define BYTEDATA_PTR(B) (GENERIC)((char*)(B) - sizeof(unsigned long))

#define PSIID(X) *(ObjectId*)BYTEDATA_DATA(X)

#define PID_FLAG "pid"
#define SID_FLAG "sid"

#define BYTE_DATA_ALLOC(A) (A *)PsiTerm::byte_data_alloc(sizeof(A))

extern ptr_definition psi_id;

class ObjFile;

class PsiTerm {

public:

/********* PSI_TO_STRING(t)
  Get the value of a Life string, or the name of a non-string psi-term.
*/
  static char* psi_to_string(ptr_psi_term t);

  /******** get_arg
    returns the argument "number" of the goal "psiTerm".
      return NULL if bad argument or missing.
	*/
  static ptr_psi_term get_arg (ptr_psi_term psiTerm, char* number);
  
  /******** get_value
    get the value of the attribute attributeName on the 
      psi-term psiTerm. returns NULL if not found
	*/
  static GENERIC get_value(ptr_psi_term psiTerm,char* attributeName,
		    ptr_psi_term& psi_value);

  static unsigned long num_of_features(ptr_node tree);

  /******** BYTE_DATA_ALLOC
    Create a bytedata block of size 
      ( sizeof(unsigned long) + bytes ) 
	*/
  static GENERIC byte_data_alloc(unsigned long bytes);
  static GENERIC byte_data_copy(GENERIC bd);

  /******** HEAP_BD_PSI_TERM
    Create a new psi_term on the heap with the given sort 
      whose value points to a bytedata block of size 
	( sizeof(unsigned long) + bytes )
	  */
  static ptr_psi_term heap_bd_psi_term(ptr_definition sort, 
				       unsigned long bytes, 
				       long stat=4);
  static ptr_psi_term heap_bd_psi_term(ptr_psi_term &, 
				       ptr_definition sort, 
				       unsigned long bytes, 
				       long stat=4);
  /******** STACK_BD_PSI_TERM
    Create a new psi_term on the stack with the given sort 
      whose value points to a bytedata block of size 
	( sizeof(unsigned long) + bytes )
	  */
  static ptr_psi_term stack_bd_psi_term(ptr_definition sort, 
				 unsigned long bytes, 
				 long stat=4);

    /******** UPDATE_SYMBOL(s)
      S is a string of characters.
      Extract the module name from S. If that fails then use the
      current module. Create the module if needed.
      Then, retrieve the keyword for 'module#symbol'. 
      Then find the correct definition by scanning the opened modules.
      */
  static ptr_definition update_symbol(ptr_module, const char* symbol); 

  /******** list_cdr
    
    */
  static ptr_psi_term list_cdr(ptr_psi_term lst);
  
  /******** list_car
    
    */
  static ptr_psi_term list_car(ptr_psi_term lst);
  
  /********* StackNil
    Create the NIL object on the stack.
      */
  static ptr_psi_term StackNil();
  
  /******** StackCons(head,tail)
    Create a CONS object.
      */
  static ptr_psi_term StackCons(ptr_psi_term head,ptr_psi_term tail);
  
  /********* StackDisjNil
    Create the DISJ_NIL object on the stack.
      */
  static ptr_psi_term StackDisjNil();
  
  
  /******** StackDisj(head,tail)
    Create a DISJUNCTION object.
      */
  static ptr_psi_term StackDisj (ptr_psi_term head, 
				      ptr_psi_term tail);

  static ptr_psi_term psi_term2psi_id(ptr_psi_term pid, 
				      ptr_psi_term sid, ObjFile* f);
  static ptr_psi_term psi_id2psi_term(ptr_psi_term id);

  static void check_psi_term(ptr_psi_term);
  static void check_psi_tree(ptr_node);
};

#endif;
