/**********************************************************

***********************************************************/

#include "OdbSource/extern-c++.h"
extern "C" {
  #include "Source/trees.h"
}

#include "rdb_channel.h"
#include "rdb.h"
#include "OdbSource/exc.h"
#include "OdbSource/psi_term.h"
#include "OdbSource/db.h"
#include "OdbSource/exc.h"

#define SYBSTRLENDECL " varchar(255) " 

/****************Externally defined variables*********/
extern ptr_definition dbnull,reg_sort;

/****************Global Varaibles********************/

static char *SqlPrefix = NULL;

static char *SelectSt = " select ";
static char *WhereSt = " where ( ";
static char *Where2St = " where not ( ";
static char *AndSt = " and ";
static char *And2St = " and not ";
static char *And3St = " ) and ( ";
static char *And4St = " ) and not ( ";
static char *CommaSt = " , ";

/****************Externally defined functions*********/

/****************Forward Declarations*****************/

/******** INT_COL(s) 
  
*/
static long int_col(char* s, char* s2)
{
  long v=1;
  char *c;

  c=s;
  if (*c==0)
    v= 0;
  else {
    for(;DIGIT(*c);c++,s2++) {
      *s2=*c;
    }
    *s2=*c;
    if (*c!=0) v= 0;
  }

  return v;
}

/******** is_reg_sort
  
*/
static long is_reg_sort(ptr_psi_term psiTerm)
{
  ptr_node nodeAttr;
  ptr_psi_term psiValue;
  
  deref_ptr(psiTerm);
  nodeAttr=find(featcmp,one,psiTerm->attr_list);
  if(nodeAttr) {  
    psiValue=(ptr_psi_term)nodeAttr->data;
    deref_ptr(psiValue);
    if(psiValue->type==reg_sort)
      return TRUE;
  }

  return FALSE;
}

/*
   Double the quotes and '#' in a character string.
*/
void RdbChannel:: safe_str(char* dest, char* src)
{
  char* s;
  char* d=dest;
  for(s=src; *s; s++) {
    *d = *s; d++;
    if (*s == '"' || *s == '#') {*d = *s; d++;}
  };
  *d='\0';
}

void RdbChannel:: quote_send(char* str, char* module_name)
{
  long quotes;  
  char *p;

  if (module_name) {
    quotes=0;
    for(p=module_name; *p; p++) {
      if (*p == '"' || *p == '#') quotes++;
    };
    if (quotes==0) append_query(module_name, " \"", "#");
    else {
      char value[strlen(module_name) + quotes + 1];
      safe_str(value, module_name);
      append_query(value, " \"", "#");
    }
  }
  quotes=0;
  for(p=str; *p; p++) {
    if (*p == '"' || *p == '#') quotes++;
  };

  if (quotes==0) {
    if (module_name) 
      append_query(str, "", "\" ");
    else
      append_query(str, " \"", "\" ");
  }
  else {
    char value[strlen(str) + quotes + 1];
    safe_str(value, str);
    if (module_name)
      append_query(value, "", "\" ");
    else
      append_query(value, " \"", "\" ");
  }
}

/******** access_tree
  
*/
void RdbChannel:: access_tree(ptr_node tree)
{
  
  if(tree) {
    if(tree->left)
      access_tree(tree->left);
    
    (sql_fn) (tree);
    
    if(tree->right)
      access_tree(tree->right);
  }

}

void RdbChannel:: ctq(ptr_node n)
{
  ptr_psi_term arg;
  char *key, buf[REAL_STRLEN];

  if (int_col((char *)n->key,buf+1)) {
    buf[0]='c';
    key=buf;
  }
  else
    key = (char *) n->key;

  arg = (ptr_psi_term) n->data;
  deref_ptr(arg);

  if (sub_type(arg->type,real)){
    append_query(key, SqlPrefix, " float ");
  }
  else    
    append_query(key, SqlPrefix, SYBSTRLENDECL);

  append_query(" NULL ","","");
  SqlPrefix = CommaSt;
}

/******** sq
  
*/
void RdbChannel:: sq(ptr_node n)
{
  char *colName;
  char *key,buf[REAL_STRLEN];

  if (int_col((char *)n->key,buf+1)) {
    buf[0]='c';
    key=buf;
  }
  else
    key = (char *) n->key;
  append_query(key, SqlPrefix," ");

  SqlPrefix = CommaSt;
}

/******** wq
  
*/
void RdbChannel:: wq(ptr_node n)
{
  ptr_psi_term arg;
  char *colName,*key;
  char buf[REAL_STRLEN];

  arg = (ptr_psi_term) n->data;

  deref_ptr(arg);

  if (arg->type != top) {
    if (int_col((char *)n->key,buf+1)) {
      buf[0]='c';
      key=buf;
    }
    else
      key = (char *) n->key;

    if (arg->value) {
      if (is_reg_sort(arg)) {
	append_query(key, SqlPrefix, " like ");
	quote_send((char *) arg->value);
      }
      else
	if (sub_type(arg->type,quoted_string)) {
	  append_query(key, SqlPrefix, " = ");
	  quote_send((char *) arg->value);
	}
	else 
	  //	  if (arg->type == real || arg->type == integer) {
	  if (sub_type(arg->type,real)) {
	    append_query(key, SqlPrefix," ");
	    sprintf(buf,"%g",*(REAL *)arg->value);
	    append_query(buf, " = ", " ");
	  }
	  else {
	    dberror << "in wq(): unknown type " 
		    << (char *)arg->type->keyword->combined_name;
	    throw_error();
	  }
    }
    else {
      if (arg->type != dbnull) {
	append_query(key, SqlPrefix, " = ");
	//	quote_send((char *)arg->type->keyword->combined_name);
	quote_send((char *)arg->type->keyword->symbol,
		   (char *)arg->type->keyword->module->module_name);
      }
      else {
	append_query(key, SqlPrefix, " is NULL ");
      }
    };
    SqlPrefix = AndSt;
  }
  
}

/******** uq
  
*/
void RdbChannel:: uq(ptr_node n)
{
  ptr_psi_term arg;
  char *colName,*key;
  char buf[REAL_STRLEN];

  arg = (ptr_psi_term) n->data;

  deref_ptr(arg);

  if (arg->type != top) {
    if (int_col((char *)n->key,buf+1)) {
      buf[0]='c';
      key=buf;
    }
    else
      key = (char *) n->key;

    if (arg->value) {
      if (sub_type(arg->type,quoted_string)) {
	append_query(key, SqlPrefix, " = ");
	quote_send((char *) arg->value);
      }
      else 
	//	  if (arg->type == real || arg->type == integer) {
	if (sub_type(arg->type,real)) {
	  append_query(key, SqlPrefix," ");
	  sprintf(buf,"%g",*(REAL *)arg->value);
	  append_query(buf, " = ", " ");
	}
	else {
	  dberror << "in uq(): unknown type " 
		  << (char *)arg->type->keyword->combined_name;
	  throw_error();
	}
    }
    else {
      append_query(key, SqlPrefix, " = ");
      //	quote_send((char *)arg->type->keyword->combined_name);
      quote_send((char *)arg->type->keyword->symbol,
		 (char *)arg->type->keyword->module->module_name);
    };
    SqlPrefix = CommaSt;
  }
  
}

/******** iq
  
*/
void RdbChannel:: iq(ptr_node n)
{
  ptr_psi_term arg;
  char buf[REAL_STRLEN];

  arg = (ptr_psi_term) n->data;
  deref_ptr(arg);

  if (arg->value) 
    if (arg->type == quoted_string) {
      append_query(SqlPrefix," "," ");
      quote_send((char *) arg->value);
    }
    else
      if (arg->type == real || arg->type == integer) {
	sprintf(buf,"%g",*(REAL *)arg->value);
	append_query(SqlPrefix," "," ");
	append_query(buf, " ", " ");
      }
      else {
	dberror << "in 1q(): unknown type " 
		    << (char *)arg->type->keyword->combined_name;
	throw_error();
      }
  else {   
      if (arg->type != dbnull) {
	append_query(SqlPrefix," "," ");
	//	quote_send((char *)arg->type->keyword->combined_name);
	quote_send((char *)arg->type->keyword->symbol,
		   (char *)arg->type->keyword->module->module_name);
      }
      else {
	append_query(" NULL ", SqlPrefix, "");
      }
  };
  SqlPrefix = CommaSt;
}

/******** create_table_query
  
*/
void RdbChannel:: create_table_query(char* relName, 
				     ptr_psi_term query)
{
  append_query(relName, " create table ", " ( ");
  deref_ptr(query);

  SqlPrefix=" ";

  sql_fn = ctq;
  access_tree((ptr_node)query->attr_list);

  append_query(" ) ", "","");

}

/******** alter_table_query
  
*/
void RdbChannel:: alter_table_query(char* relName, 
				     ptr_psi_term query)
{
  append_query(relName, " alter table ", " add ");
  deref_ptr(query);

  SqlPrefix=" ";

  sql_fn = ctq;
  access_tree((ptr_node)query->attr_list);

  //  append_query("  ", "","");

}

/******** insert_query
  
*/
void RdbChannel:: insert_query(char* relName, ptr_psi_term query)
{

  append_query(relName, " insert into ", " ( ");
  deref_ptr(query);

  SqlPrefix=" ";

  sql_fn = sq;
  access_tree((ptr_node)query->attr_list);

  SqlPrefix=" ) values ( ";

  sql_fn=iq;
  access_tree((ptr_node)query->attr_list);

  append_query(" ) ", "","");

}

/******** update_query
  
*/
void RdbChannel:: update_query(char* relName, ptr_psi_term query)
{

  append_query(relName, " update ", " set ");
  deref_ptr(query);

  SqlPrefix=" ";

  sql_fn = uq;
  access_tree((ptr_node)query->attr_list);
}

/******** select_query
  
*/
void RdbChannel:: select_query(char* relName,
			       ptr_psi_term query,
			       ptr_psi_term notQuery)
{
  ptr_psi_term car,cdr;

  deref_ptr(query);
  deref_ptr(notQuery);

  SqlPrefix = SelectSt;

  sql_fn = sq;
  access_tree((ptr_node)query->attr_list);

  append_query(relName, " from ", " ");
  
  SqlPrefix = WhereSt;

  sql_fn=wq;
  access_tree((ptr_node)query->attr_list);

  if (SqlPrefix == WhereSt) 
    SqlPrefix = Where2St;
  else
    SqlPrefix = And4St;
  if (notQuery->type==alist) {
    cdr = notQuery;
    sql_fn=wq;
    while (cdr->type != nil) {
      car = PsiTerm::list_car(cdr);
      access_tree((ptr_node)car->attr_list);
      cdr = PsiTerm::list_cdr(cdr);
      if (SqlPrefix != Where2St) 
	SqlPrefix = And4St;
    }
  }
  
  if ( SqlPrefix != WhereSt && SqlPrefix != Where2St)
    append_query(")", " ", " ");
 
}
