/**********************************************************

	rdb_built_ins.c
	Database functions and predicates for
	the wild_LIFE interpreter v1.01

***********************************************************/

#include <string.h>

#include "rdb_channel.h"
#include "OdbSource/extern-c++.h"
#include "sql_psi.h"
#include "OdbSource/dbtemplate.h"
#include "sybase.h"

extern "C" {

void check_psi_term(ptr_psi_term*);
void check_definition(ptr_definition*);

}

extern void insert_misc_builtins();

#define DBCHANNEL(X) *(RdbChannel**)BYTEDATA_DATA(X)
#define STRNG(X) (char *)val[X]
#define INT(X) (long)*(REAL *)val[X]

/****************Externally defined variables*********/
extern S2P_Value s2p_val;

/****************Global Varaibles********************/
int error_off=0;
ptr_definition rdb_channel,dbnull,reg_sort;

static char *feature[] = {"1","2","3","4","5"};

/****************Externally defined functions*********/

/****************Forward Declarations*****************/


/********************* RDB_OPEN_STREAM ***********************/
/*	C_RDB_OPEN_STREAM -- LIFE pred: rdb_open_stream
	rdb_open_stream(+DatabaseName,+UserPwd,+Server,-DbChannel)
*/
static long c_rdb_open_stream()
{
  declare_var(4);

  types[0] = quoted_string;
  types[1] = quoted_string;
  types[2] = quoted_string;

  begin_dbbuiltin(4,3);


  if (is_top(args[3])) {

    ptr_psi_term psi_db_channel = 
      PsiTerm::heap_bd_psi_term( args[3], rdb_channel, 
				sizeof(RdbChannel*) );
    GENERIC bd = psi_db_channel->value;
//    DBCHANNEL(bd) = new RdbChannel(STRNG(0), STRNG(1), 
//				   STRNG(2));
    DBCHANNEL(bd) = new Sybase(STRNG(0), STRNG(1), 
			       STRNG(2));

  }
  else {
    Errorline("bad stream in %P.\n",psiTerm);
    success=false;
  }

  end_dbbuiltin();
  
}

/********************* rdb_close_stream ***********************/
/*	c_rdb_close_stream -- LIFE pred: rdb_close_stream
	rdb_close_stream(+DbChannel)
*/
static long c_rdb_close_stream()
{
  declare_var(1);

  types[0] = rdb_channel;

  begin_dbbuiltin(1,1);
      
  RdbChannel* ch = DBCHANNEL(val[0]);
  delete ch;

  end_dbbuiltin();
}

/********************* rdb_get_dbname ***********************/
/*      c_rdb_get_dbname -- LIFE pred: rdb_get_dbname
        rdb_get_dbname(+DbChannel)
*/
static long c_rdb_get_dbname()
{
  declare_var(1);
  
  types[0] = rdb_channel;

  begin_dbbuiltin(1,1);

  ptr_psi_term res = aim->b;
  deref(res);
  RdbChannel* ch = DBCHANNEL(val[0]);
  char* name = ch->db_name();
  
  ptr_psi_term t1 = stack_psi_term(4);
  t1->type=quoted_string;
  char* dbname = (char*)heap_alloc(strlen(name)+1);
  strcpy(dbname, name);
  t1->value=(GENERIC)dbname;
  push_goal(unify, t1, res, NULL);
  end_dbbuiltin();
}

/********************* rdb_error_mode ***********************/
/*	c_rdb_error_mode -- LIFE pred: rdb_error_mode
	rdb_error_mode(+A:int)
	A=1: Turn off error reporting
	A=0: Turn on error reporting
*/
static long c_rdb_error_mode()
{
  declare_var(1);

  types[0] = integer;

  begin_dbbuiltin(1,1);
  error_off = (INT(0)>0)?1:0;
  end_dbbuiltin();
}

/********************* RDB_SELECT ***********************/
/*	C_RDB_SELECT -- LIFE pred: rdb_select
	rdb_select(+DbChannel, +Table, +Query, +NotQuery)
	Query = @(ColName => @/Value),
	ColName is inserted into select statement, its value is
	put in "where statement" unless it is top
	NotQuery = [ @(ColName => Value), ...]
*/
static long c_rdb_select()
{
  declare_var(4);

  types[0] = rdb_channel;
  types[1] = quoted_string;

  begin_dbbuiltin(4,2);
  
  RdbChannel* ch = DBCHANNEL(val[0]);
  ch->select_query(STRNG(1), args[2], args[3]);
  
  end_dbbuiltin();
}

/********************* RDB_INSERT ***********************/
/*	C_RDB_INSERT -- LIFE pred: rdb_insert
	rdb_insert(+DbChannel, +Table, +Query)
	Query = @(ColName => Value)
*/
static long c_rdb_insert()
{
  declare_var(3);

  types[0] = rdb_channel;
  types[1] = quoted_string;

  begin_dbbuiltin(3,2);
  
  if (args[2] -> attr_list) {
    RdbChannel* ch = DBCHANNEL(val[0]);
    ch->insert_query(STRNG(1), args[2]);
  }
  else {
    Errorline ("bad arguments in %P.\n", psiTerm); 
    success=FALSE;
  }

  end_dbbuiltin();
}

/********************* RDB_UPDATE ***********************/
/*	C_RDB_UPDATE -- LIFE pred: rdb_update
	rdb_update(+DbChannel, +Table, +Query)
	Query = @(ColName => Value)
*/
static long c_rdb_update()
{
  declare_var(3);

  types[0] = rdb_channel;
  types[1] = quoted_string;

  begin_dbbuiltin(3,2);
  
  if (args[2] -> attr_list) {
    RdbChannel* ch = DBCHANNEL(val[0]);
    ch->update_query(STRNG(1), args[2]);
  }
  else {
    Errorline ("bad arguments in %P.\n", psiTerm); 
    success=FALSE;
  }

  end_dbbuiltin();
}

/********************* RDB_CREATE_TABLE ***********************/
/*	C_RDB_CREATE_TABLE -- LIFE pred: rdb_create_table
	rdb_create_table(+DbChannel, +Table, +Query)
	Query = @(ColName => X)
	if X is of type real then float column is 
	created else string column.
*/
static long c_rdb_create_table()
{
  declare_var(3);

  types[0] = rdb_channel;
  types[1] = quoted_string;

  begin_dbbuiltin(3,2);
  
  if (args[2] -> attr_list) {
    RdbChannel* ch = DBCHANNEL(val[0]);
    ch->create_table_query(STRNG(1), args[2]);
  }
  else {
    Errorline ("bad arguments in %P.\n", psiTerm); 
    success=FALSE;
  }

  end_dbbuiltin();
}


/********************* RDB_ALTER_TABLE ***********************/
/*	C_RDB_ALTER_TABLE -- LIFE pred: rdb_alter_table
	rdb_alter_table(+DbChannel, +Table, +Query)
	Query = @(ColName => X)
	if X is of type real then float column is 
	created else string column.
*/
static long c_rdb_alter_table()
{
  declare_var(3);

  types[0] = rdb_channel;
  types[1] = quoted_string;

  begin_dbbuiltin(3,2);
  
  if (args[2] -> attr_list) {
    RdbChannel* ch = DBCHANNEL(val[0]);
    ch->alter_table_query(STRNG(1), args[2]);
  }
  else {
    Errorline ("bad arguments in %P.\n", psiTerm); 
    success=FALSE;
  }

  end_dbbuiltin();
}

/********************* rdb_sql ***********************/
/*	C_rdb_sql -- LIFE pred: rdb_sql
	rdb_sql(+DbChannel, +Query)
*/
static long c_rdb_sql()
{
  declare_var(2);

  types[0] = rdb_channel;
  types[1] = quoted_string;

  begin_dbbuiltin(2,2);
  
  RdbChannel* ch = DBCHANNEL(val[0]);
  ch->append_query(STRNG(1), " ", " ");

  end_dbbuiltin();
}

/********************* rdb_sql_send ***********************/
/*	c_rdb_sql_send -- LIFE pred: rdb_sql_send
	rdb_sql_send(+DbChannel)
*/
static long c_rdb_sql_send()
{
  declare_var(1);

  types[0] = rdb_channel;

  begin_dbbuiltin(1,1);
  
  RdbChannel* ch = DBCHANNEL(val[0]);
  ch->sql_send();

  end_dbbuiltin();
}

/********************* rdb_sql_ok ***********************/
/*	c_rdb_sql_ok -- LIFE pred: rdb_sql_ok
	rdb_sql_ok(+DbChannel)
*/
static long c_rdb_sql_ok()
{
  declare_var(1);

  types[0] = rdb_channel;

  begin_dbbuiltin(1,1);
  
  RdbChannel* ch = DBCHANNEL(val[0]);
  ch->sql_ok();

  end_dbbuiltin();
}

/********************* rdb_execute ***********************/
/*	c_dbresult -- LIFE pred: dbresult
	rdb_execute(+DbChannel)
*/
static long c_rdb_execute()
{
  int hasResults, hasRows;
  declare_var(1);

  types[0] = rdb_channel;

  begin_dbbuiltin(1,1);

  ptr_psi_term result = aim->b;
  deref(result);

  RdbChannel* ch = DBCHANNEL(val[0]);
  //  s2p_val.init();
  int what = ch->execute();
  if (what==0)
    success=FALSE;
  else {
    ptr_psi_term term = stack_psi_term(4);
    if (what==2) // has returned sql data to be processed further
      term->type=true_type;
    else 
      term->type=false_type;
    resid_aim=NULL;
    push_goal(unify, term, result, NULL);
  }
  end_dbbuiltin();

}

/********************* rdb_data ***********************/
/*	c_rdb_data -- LIFE pred: rdb_data
	rdb_data(+DbChannel, -Result)
	
*/
static long c_rdb_data()
{
  declare_var(2);

  types[0] = rdb_channel;

  begin_dbbuiltin(2,1);
  
  //  QueryPsiTerm=Db_StackDisjNil();
  RdbChannel* ch = DBCHANNEL(val[0]);
  if(success=ch->data())
    push_goal(unify, args[1], s2p_val.result());

  end_dbbuiltin();
}

/********************* rdb_cancel ***********************/
/*	c_rdb_cancel -- LIFE pred: rdb_cancel
	rdb_cancel(+DbChannel)
	can be called only after calling rdb_sql_send
*/
static long c_rdb_cancel()
{
  declare_var(2);

  types[0] = rdb_channel;

  begin_dbbuiltin(1,1);
  
  RdbChannel* ch = DBCHANNEL(val[0]);
  ch->cancel_all();

  end_dbbuiltin();
}

/********************* rdb_cancel_query ***********************/
/*	c_rdb_cancel_query -- LIFE pred: rdb_cancel_query
	rdb_cancel_query(+DbChannel)
	can be called only after calling dbresult
*/
static long c_rdb_cancel_query()
{
  declare_var(2);

  types[0] = rdb_channel;

  begin_dbbuiltin(1,1);
  
  RdbChannel* ch = DBCHANNEL(val[0]);
  ch->cancel_query();

  end_dbbuiltin();
}

/********************* rdb_use ***********************/
/*	c_rdb_use -- LIFE pred: rdb_use
	rdb_use(+DbChannel)
*/
static long c_rdb_use()
{
  declare_var(2);

  types[0] = rdb_channel;
 
  begin_dbbuiltin(1,1);

  RdbChannel* ch = DBCHANNEL(val[0]);
  ch->database(STRNG(1));

  end_dbbuiltin();
}

/********************* RDB_PRINT ***********************/
/*	C_RDB_PRINT -- LIFE pred: rdb_print
	rdb_print(+DbChannel)
*/
static long c_rdb_print()
{
  declare_var(1);

  types[0] = rdb_channel;

  begin_dbbuiltin(1,1);
  
  RdbChannel* ch = DBCHANNEL(val[0]);
  cout << *ch << "\n";
  end_dbbuiltin();
}

static long c_rdb_reg_convert()
{
  declare_var(1);
  types[0] = quoted_string;
  begin_dbbuiltin(1,1);

  ptr_psi_term result=aim->b;
  deref(result);

  char* reg_str1= STRNG(0);

  int count = 0; 
  char *p;
  for(p=reg_str1; *p; p++) {
    if (*p == '%' || *p == '.' || *p == '*' || *p == '\\') 
      count++;
  };

  char* reg_str2 = (char *) heap_alloc(strlen(reg_str1)+3+count*2);
  char* p2=reg_str2;

  *(p2++) = '^';  
  for (p=reg_str1; *p; p++,p2++) {

    switch (*p) {

    case '%':
      *(p2++) = '.';
      *p2 = '*';
      break;

    case '_':
      *p2 = '.';
      break;
      
    case '[':
      *p2 = *p;
      if ( *(p+1) == ']' )
	/*	*(++p2) = *(++p);
	 */
	{
	  Errorline("An empty string of characters in square brackets not allowed: \"%s\"\n", reg_str1);
	  success=FALSE;
	};
      while (*p && *p != ']')
	*(++p2) = *(++p);
      break;

    case '\\':
    case '.':
    case '*': 
      *(p2++) = '\\';
    default:
      *p2 = *p;
      
    }
  
  };
  *(p2++) = '$';
  *p2=0;
  
  ptr_psi_term arg3=stack_psi_term(0);
  arg3->type=quoted_string;
  arg3->value=(GENERIC) reg_str2;
    
  push_goal(unify,arg3,result,NULL);

  end_dbbuiltin();
}

void check_rdb_definitions() {
  check_definition(&rdb_channel);
  check_definition(&dbnull);
  check_definition(&reg_sort);
}

void init_rdb_builtins()
{
  ptr_module curmod = current_module;
  set_current_module(rdb_module);

  rdb_channel = update_symbol(rdb_module,"rdb_channel");
  reg_sort = update_symbol(rdb_module,"reg_sort");
  dbnull = update_symbol(rdb_module,"NULL");

  new_built_in(rdb_module,"rdb_open_stream",predicate,c_rdb_open_stream);
  new_built_in(rdb_module,"rdb_close_stream",predicate,c_rdb_close_stream);
  new_built_in(rdb_module,"rdb_get_dbname",function,c_rdb_get_dbname);
  new_built_in(rdb_module,"rdb_error_mode",predicate,c_rdb_error_mode);

  new_built_in(rdb_module,"rdb_select",predicate,c_rdb_select);
  new_built_in(rdb_module,"rdb_create_table",predicate,c_rdb_create_table);
  new_built_in(rdb_module,"rdb_alter_table",predicate,c_rdb_alter_table);
  new_built_in(rdb_module,"rdb_insert",predicate,c_rdb_insert);
  new_built_in(rdb_module,"rdb_update",predicate,c_rdb_update);
  new_built_in(rdb_module,"rdb_sql",predicate,c_rdb_sql);

  new_built_in(rdb_module,"rdb_sql_send",predicate,c_rdb_sql_send);
  new_built_in(rdb_module,"rdb_sql_ok",predicate,c_rdb_sql_ok);
  new_built_in(rdb_module,"rdb_execute",function,c_rdb_execute);
  new_built_in(rdb_module,"rdb_data",predicate,c_rdb_data);
  new_built_in(rdb_module,"rdb_cancel",predicate,c_rdb_cancel);
  new_built_in(rdb_module,"rdb_cancel_query",predicate,c_rdb_cancel_query);

  new_built_in(rdb_module,"rdb_print",predicate,c_rdb_print);
  new_built_in(rdb_module,"rdb_reg_convert",function,c_rdb_reg_convert);

  new_built_in(rdb_module,"rdb_use",predicate,c_rdb_use);
  set_current_module(curmod);
}

