/**********************************************************

        sybase.cpp:
	Database functions for accessing SYBASE

***********************************************************/

/******* useful sybase database functions
  dbsetinterrupt, dbsetbusy, dbsetidle, 
  dbsetlogintime, dbsafestr, dbsetifile, dbuse, dbsetopt,
  DBMORECMDS, DBCOUNT, DBCURROW
  dbfreebuf, dbcancel, dbcanquery, dbclrbuf
  dbprrow, dbprhead,
  dbname, dbexit, dbclose, DBSETLENCRYPT, dbsetnull, dbcursor
*/


#include "rdb.h"
#include "sybase.h"
#include "OdbSource/exc.h"
#include "OdbSource/crypt.h"
#include "sql_psi.h"
#include "OdbSource/exc.h"

extern "C" {
//#include <syberror.h>  
void install_syberr();
}

#define REAL double

/****************Externally defined variables*********/
S2P_Value s2p_val;

/****************Global Varaibles********************/

/****************Externally defined functions*********/


Sybase:: Sybase(char* user, char* user_pwd, char* server):
                          db_stream(NULL) 
{
  init();
  connect(user, user_pwd, server);
}

Sybase:: ~Sybase() {
  close();
}

void Sybase:: init()
{
  if (dbinit() == FAIL) {
    dberror << "Could not initialize the database";
    throw_error();
  }
  /* Install  error-handling and message-handling
   * routines. 
   */
//  dberrhandle(syberr_handler);
//  dbmsghandle(sybmsg_handler);
  install_syberr();
}

void Sybase:: connect(char* user, char* user_pwd, char* server)
{

  LOGINREC *login;

  if(!(login=dblogin())) {
    dberror << "Could not allocate database login record structure";
    throw_error();
  }

  DBSETLUSER(login, user);
  DBSETLPWD(login, (const char*)Encode(user_pwd));
  
  /*
   ** Get a DBPROCESS structure for communicating with SQL Server.
   ** A NULL servername defaults to the server specified by DSQUERY.
   */

  if ( !(db_stream = dbopen(login, server)) ) {
    dberror << "Could not connect to the database server";
    throw_error();
  }

  /*
   ** LOGINREC structure no longer required after dbopen() call
   */
  dbloginfree(login);
  
}

/******** close
  Close and de-allocate DBPROCESS structure
*/
void Sybase:: close()
{

  dbclose( db_stream );

}

/******** db_name
  Return the name of the current database.
*/
char* Sybase:: db_name()
{
  return dbname( db_stream );
}

/******** database
  Use database db.
*/
void Sybase:: database(char* db)
{

  if ( ! dbuse(db_stream,db) ) {
    throw_error();
  }
}

/******** sql_send
  dbsqlsend Send a command batch  to  the  server  and  do  not  wait  for  a
  response.
*/
void Sybase:: sql_send()
{
  if ( ! dbsqlsend(db_stream) ){
    throw_error();
  }
    

}

/******** cancel_query
  dbcanquery: Cancel any rows pending from the most recently executed query.
  can be called only after calling dbresult().
*/
void Sybase:: cancel_query()
{
  if ( !  dbcanquery( db_stream ) ){
    throw_error();
  }
   
}

/******** cancel_all
  dbcancel can be called only after calling dbsqlsend(); this cancels 
  all the commands and  flushes any pending results.
*/
void Sybase:: cancel_all()
{
  if ( !  dbcancel( db_stream ) ){
    throw_error();
  }
}

/******** sql_ok
   dbsqlok: Wait for results from the server and verify  the  correctness  of
   the instructions the server is responding to.
   The application must call dbsqlok() after a call to dbsqlsend()
   or dbrpcsend(), and before and  after  a  set  of  dbmoretext()
   calls.   A successful dbsqlok() call must always be followed by
   a call to dbresults() to process the results.
*/
void Sybase:: sql_ok()
{

  if ( ! dbsqlok(db_stream) ){
    throw_error();
  }
}


/******** execute
  dbresults: routine sets up the next command in the command batch  for
  processing.  The application program calls it after dbsqlexec()
  or dbsqlok() returns SUCCEED.  The first  call  to  dbresults()
  will  always  return  either  SUCCEED or NO_MORE_RESULTS if the
  call to dbsqlexec() or dbsqlok() has  returned  SUCCEED.  
*/
long Sybase:: execute()
{
  int i;
  RETCODE ret_code;

  while (1) {
    ret_code = dbresults(db_stream);
    if (ret_code == SUCCEED ) {
      if ( DBROWS(db_stream) ) {
	s2p_val.init();
	s2p_val.num_of_columns = dbnumcols(db_stream);	
	int num_of_columns=s2p_val.num_of_columns;
	s2p_val.tuple = new Attribute[num_of_columns];
	Attribute* tuple=s2p_val.tuple;
	int sql_type;
	for (i=0; i < num_of_columns; i++) {
	  sql_type = tuple[i].sql_type = dbcoltype(db_stream, i+1);
	  if (sql_type == SYBCHAR) 
	    tuple[i].type = DB_STRTYPE;
	  else 
	    if (sql_type == SYBFLT8 || sql_type == SYBINT4 
		|| sql_type == SYBREAL || sql_type == SYBBIT 
		|| sql_type == SYBINT2 || sql_type == SYBINT1)
	      tuple[i].type = DB_NUMTYPE;
	    else {
	      dberror << "in Sybase:: data(): unknown type";
	      throw_error();
	    };
	  tuple[i].col_name = dbcolname(db_stream, i+1);	  
	}
	return 2;
	//	break;
      }
    }
    else
      if ( ret_code == FAIL ){
	//	success=FALSE;
	//	break;
	return 0;
      }
      else {/* ret_code = NO_MORE_RESULTS */
	//	break;
	return 1;
      }
  }
}

/******** data
  
*/
long Sybase:: data()
{
  int i;
  STATUS st;
  int num_of_columns=s2p_val.num_of_columns;
  Attribute* tuple=s2p_val.tuple;

  if ( (st=dbnextrow(db_stream)) == REG_ROW) {
    for (i=0; i < num_of_columns; i++) {
      
      tuple[i].data = (char *)dbdata(db_stream,i+1);
      // dbdata doesnt return null terminted strings, see dbdatlen
      // for string length
      tuple[i].length = dbdatlen(db_stream, i+1);
      
      if ( tuple[i].type == DB_STRTYPE) {
      }
      else { /* tuple[i].type == DB_NUMTYPE */
	if ( tuple[i].length != 0 ) {
	  tuple[i].length = sizeof(REAL);
	  switch (tuple[i].sql_type)
	    {
	    case SYBFLT8:
	      tuple[i].data2 = (REAL) ( *(DBFLT8 *)tuple[i].data );
	      break;
	      
	    case SYBINT4:
	      tuple[i].data2 = (REAL) ( *(DBINT *)tuple[i].data );
	      break;
	      
	    case SYBREAL:
	      tuple[i].data2 = (REAL) ( *(DBREAL *)tuple[i].data );
	      break;
	      
	    case SYBBIT:
	      tuple[i].data2 = (REAL) ( *(DBBIT *)tuple[i].data );
	      break;
	      
	    case SYBINT2:
	      tuple[i].data2 = 
		(REAL) ( *(DBSMALLINT *)tuple[i].data );
	      break;
	      
	    case SYBINT1:
	      tuple[i].data2 = 
		(REAL) ( *(DBTINYINT *)tuple[i].data );
	      break;
	      
	    default:
	      /*error*/
	      break;
	    };
	  tuple[i].data = (char *) &tuple[i].data2;
	}
      }
    }
    s2p_val.tuple2psi(tuple, num_of_columns); 
    /*should exit_life if fail*/
    return TRUE;
  }
  else {
    return FALSE;
  }
}

/******** append_query
  Add sql text to the DBPROCESS command buffer 
*/
void Sybase:: append_query(char* query, char* prefix, char* suffix)
{

  if ( ! dbfcmd( db_stream, "%s%s%s", prefix, query, suffix ) ) {
    throw_error();
  }
}

void Sybase:: output_data(ostream& os) const
{
  int cmd_buf_len;
  long success = TRUE;

  cmd_buf_len = dbstrlen(db_stream);

  char dest[cmd_buf_len+1];
//    (char *) malloc( (cmd_buf_len+1)*sizeof(char) );
  if ( dbstrcpy(db_stream, 0, cmd_buf_len, dest) ) {
//    printf("%s\n",dest);
    os << dest;
  }
}

/******** db_exit
  Close and de-allocate all DBPROCESS structures, and clean up  any
  structures initialized by dbinit().
*/
void Sybase:: db_exit()
{

  dbexit();

}

